/* @(#)smedit.c 2.83.2.2 Software Membership 2/17/96 */
/*
	SM/Editor - Version 2

    ---------
    Copyright (C) Si-Won Cho
            Dept. of Electrical Engineering
            Dongguk University
            Seoul, KOREA
            chosw@dgu4680.dongguk.ac.kr

            SAMSUNG Electronics
            Software Membership

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    ---------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef hpux
#include <strings.h>
#endif
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <locale.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/MainW.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/SeparatoG.h>
#include <Xm/SelectioB.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/ToggleBG.h>
#if (XmVersion >= 1002)	/* Motif 1.2 */
#include <Xm/RepType.h>
#include <Xm/DragDrop.h>
#endif
#include "pattern.h"
#include "list.h"
#include "strctrl.h"
#include "help.h"
#define	ABOUT_STR  "Version 2.83.2.2\n\n\
Copyright (C) Si-Won Cho\n\
Software Membership\n\
Seoul, KOREA\n\
chosw@dgu4680.dongguk.ac.kr"

#define	WARNING_NEW	0
#define	WARNING_OPEN	1
#define WARNING_LIST	2
#define	WARNING_QUIT	3

#define	SAVE_DEFAULT	0
#define	SAVE_NEW	1
#define	SAVE_OPEN	2
#define	SAVE_LIST	3
#define	SAVE_QUIT	4

#define FILE_SAVE	0
#define FILE_OPEN	1

#define MAXTAB		80
#define MAXLIST		10

/* external variable declarations */
extern char *sys_errlist[];
extern int errno;

/*	Global Widgets		*/
Widget	toplevel;
Widget	main_w, menubar, form, separator;
Widget	file_menu, edit_menu, search_menu, option_menu, help_menu;
Widget	undo_button, redo_button;
Widget	popup_undo_button, popup_redo_button;
Widget	pick_submenu, pick_button[MAXLIST], option_env, option_pref;
Widget	popup_menu, popup_submenu;
Widget	text_w;
Widget	form_status, line_label, line_text, col_label, col_text;
Widget	status_line, env_line_1, env_line_2;
Widget	OpenDialog, SaveAsDialog;
Widget	OverwriteDialog, QuestionDialog, ErrorDialog;
Widget	go_dialog;
Widget	SearchDialog;
Widget	tab_dialog, block_dialog;
Widget	help_dialog, about_dialog;
Widget	ScrollBar;

/*	Global Variables	*/
static	LinkPtr	undo = (LinkPtr)NULL, redo = (LinkPtr)NULL;
char	*FileName;
char 	*OpenName = NULL;
Boolean file_saved = True;		/* Editing File Modified     */
XmTextPosition last_position;		/* the last cursor Position  */
static	XmTextPosition place_marker[10];	/* Place Mark Position */
Boolean create_backup = True;		/* Create BackUp File   */
Boolean	indent_mode = True;		/* Toggle Indent Mode	*/
Boolean	fill_mode = True;		/* Toggle Tab Mode	*/
Boolean optimize_text = True;           /* Optimize Text Mode   */
Boolean show_cursor_position = True;	/* Display Line Column	*/
Boolean show_match = False;	        /* Auto Matching	*/
Boolean insert_mode = True;	        /* Toggle Insert Mode   */
Boolean undo_after_save = True;		/* Undo After Save 	*/
Boolean pending_delete = True;		/* Pending delete Mode	*/
Boolean ignoreModify = False;		/* Ignore Modify	*/
int	TabSize = 8;			/* Tab Size		*/
int	BlockIndentSize = 2;		/* Block Indent		*/
int	SaveNum;			/* case of Save As...	*/
int	WarningNum;			/* case of Waring DialogBox */
int 	file_count = 0;			/* No. of  file list */
static XmTextScanType selectionArray[] = {
        XmSELECT_POSITION, XmSELECT_ALL
};

/*	Global Functions	*/
#ifdef _NO_PROTO
void	usage();
void    update_StatusLine();
void	add_name();
void	set_title();
Boolean open_file();
Boolean save_file();
#else		/* _NO_PROTO */
void	usage(char *);
void    update_StatusLine(Widget, XmTextPosition);
void	add_name(char *);
void	set_title(char *);
Boolean open_file(char *);
Boolean save_file(char *);
#endif

void    update_EnvironmentLine();
void	load_option();

#ifdef _NO_PROTO
Widget	CreateMenuBar();
#else		/*_NO_PROTO */
Widget	CreateMenuBar(Widget);
#endif

/*	XtEventHandler	*/
#ifdef _NO_PROTO
void	popup_cb();
#else		/* _NO_PROTO */
void	popup_cb(Widget, XtPointer, XButtonPressedEvent *, Boolean *);
#endif

/*      MenuBar Callback Proc   */
#ifdef _NO_PROTO
void 	FileMenu_cb();
void    EditMenu_cb();
void    SearchMenu_cb();
void    OptionMenu_cb();
void    EnvMenu_cb();
void    PrefMenu_cb();
void    PopupMenu_cb();
void    PopupSubMenu_cb();
void    HelpMenu_cb();
#else		/*_NO_PROTO */
void 	FileMenu_cb(Widget, XtPointer, XtPointer);
void    EditMenu_cb(Widget, XtPointer, XtPointer);
void    SearchMenu_cb(Widget, XtPointer, XtPointer);
void    OptionMenu_cb(Widget, XtPointer, XtPointer);
void    EnvMenu_cb(Widget, XtPointer, XtPointer);
void    PrefMenu_cb(Widget, XtPointer, XtPointer);
void    PopupMenu_cb(Widget, XtPointer, XtPointer);
void    PopupSubMenu_cb(Widget, XtPointer, XtPointer);
void    HelpMenu_cb(Widget, XtPointer, XtPointer);
#endif

#ifdef _NO_PROTO
void	pick_cb();
#else		/* _NO_PROTO */
void	pick_cb(Widget, XtPointer, XtPointer);
#endif

/*      Widget Callback Proc	*/
#ifdef _NO_PROTO
void    CloseBox();
void	CloseSelectionBox();
void	Modify_cb();
void    Motion_cb();
void	open_cb();
void	SaveAs_cb();
void    OpenDir_cb();
void    OverWrite_cb();
void    yes_cb();
void    no_cb();
void    go_cb();
void    settab_cb();
void    setblock_cb();
#else		/* _NO_PROTO */
void    CloseBox(Widget, XtPointer, XtPointer);
void    CloseSelectionBox(Widget, XtPointer, XtPointer);
void	Modify_cb(Widget, XtPointer, XtPointer);
void    Motion_cb(Widget, XtPointer, XtPointer);
void	open_cb(Widget, XtPointer, XtPointer);
void    SaveAs_cb(Widget, XtPointer, XtPointer);
void    OpenDir_cb(Widget, XtPointer, XtPointer);
void    OverWrite_cb(Widget, XtPointer, XtPointer);
void    yes_cb(Widget, XtPointer, XtPointer);
void    no_cb(Widget, XtPointer, XtPointer);
void    go_cb(Widget, XtPointer, XtPointer);
void    settab_cb(Widget, XtPointer, XtPointer);
void    setblock_cb(Widget, XtPointer, XtPointer);
#endif

/*	Translation Action Proc */
#ifdef _NO_PROTO
void 	UndoLast();
void 	UndoAll();
void	RedoAction();
void 	process_Ctrl_K();
void 	process_Ctrl_Q();
void 	process_Ctrl_O();
void 	process_tab();
void    BlockIndent();
void    BlockUnindent();
void 	move_top();
void 	move_bottom();
void 	delete_line();
void 	delete_next_word();
void 	delete_previous_word();
void 	word_left();
void 	word_right();
void 	show_info();
#else		/* _NO_PROTO */
void 	UndoLast(Widget, XEvent *, String *, Cardinal *);
void 	UndoAll(Widget, XEvent *, String *, Cardinal *);
void 	RedoAction(Widget, XEvent *, String *, Cardinal *);
void 	process_Ctrl_K(Widget, XEvent *, String *, Cardinal *);
void 	process_Ctrl_Q(Widget, XEvent *, String *, Cardinal *);
void 	process_Ctrl_O(Widget, XEvent *, String *, Cardinal *);
void 	process_tab(Widget, XEvent *, String *, Cardinal *);
void 	BlockIndent(Widget, XEvent *, String *, Cardinal *);
void 	BlockUnindent(Widget, XEvent *, String *, Cardinal *);
void 	move_top(Widget, XEvent *, String *, Cardinal *);
void 	move_bottom(Widget, XEvent *, String *, Cardinal *);
void 	delete_line(Widget, XEvent *, String *, Cardinal *);
void 	delete_next_word(Widget, XEvent *, String *, Cardinal *);
void 	delete_previous_word(Widget, XEvent *, String *, Cardinal *);
void 	word_left(Widget, XEvent *, String *, Cardinal *);
void 	word_right(Widget, XEvent *, String *, Cardinal *);
void 	show_info(Widget, XEvent *, String *, Cardinal *);
#endif

/*	Translation Action Table		*/
static XtActionsRec text_actions[] = {
	{"replace-find-text", (XtActionProc) ReplaceFindText},
	{"UndoLast", (XtActionProc) UndoLast},
	{"UndoAll", (XtActionProc) UndoAll},
	{"RedoAction", (XtActionProc) RedoAction},
	{"process-Ctrl-K", (XtActionProc) process_Ctrl_K},
	{"process-Ctrl-Q", (XtActionProc) process_Ctrl_Q},
	{"process-Ctrl-O", (XtActionProc) process_Ctrl_O},
	{"Process-tab", (XtActionProc) process_tab},
	{"block-indent", (XtActionProc) BlockIndent},
	{"block-unindent", (XtActionProc) BlockUnindent},
	{"move-top", (XtActionProc) move_top},
	{"move-bottom", (XtActionProc) move_bottom},
	{"delete-line", (XtActionProc) delete_line},
	{"Delete-next-word", (XtActionProc) delete_next_word},
	{"Delete-previous-word", (XtActionProc) delete_previous_word},
	{"word-left", (XtActionProc) word_left},
	{"word-right", (XtActionProc) word_right},
	{"show-info", (XtActionProc) show_info},
	{NULL, NULL}
};

static char *set_unindent = "#override \n\
	<Key>Return:		newline()\n\
";

static char *set_indent = "#override \n\
	<Key>Return:		newline-and-indent()\n\
";

static char *optimal_fill = "#override \n\
	<Key>Tab:		process-tab()\n\
";

static char *normal_fill = "#override \n\
	<Key>Tab:		Process-tab()\n\
";

/*
static XrmOptionDescRec options[] = {
        {"-undo",       "*option_env*button_3.set", XrmoptionNoArg, "True"},
        {"+undo",       "*option_env*button_3.set", XrmoptionNoArg, "False"},
        {"-showmatch",  "*option_env*button_4.set", XrmoptionNoArg, "True"},
        {"+showmatch",  "*option_env*button_4.set", XrmoptionNoArg, "False"},
        {"-backup",     "*option_pref*button_0.set", XrmoptionNoArg, "True"},
        {"+backup",     "*option_pref*button_0.set", XrmoptionNoArg, "False"},
        {"-rw",         "*option_pref*button_1.set", XrmoptionNoArg, "True"},
        {"+rw",         "*option_pref*button_1.set", XrmoptionNoArg, "False"}
};
*/

#ifdef _NO_PROTO
void main(argc, argv)
int argc;
char *argv[];
#else		/* _NO_PROTO */
void main(int argc, char *argv[])
#endif
{
	XtAppContext app_context;
	Arg	args[9];
	XmString str;
	char *buf;
	int i, line;
	Boolean is_Empty;

	load_option();		/* Load Option data file */

	XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);
#if (XmVersion >= 1002)	/* Motif 1.2 */
	XmRepTypeInstallTearOffModelConverter();
#endif

	toplevel = XtAppInitialize(&app_context, "SMedit",
			(XrmOptionDescList)NULL, 0,
/*			options, XtMumber(options),*/
			&argc, argv,
			(String *)NULL, (ArgList)NULL, 0);

	main_w = XtCreateWidget("main_w",
		xmMainWindowWidgetClass, toplevel, (ArgList)NULL, 0);

	menubar = CreateMenuBar(main_w);

	XtManageChild(menubar);

        form = XtCreateWidget("form",
		xmFormWidgetClass, main_w, (ArgList)NULL, 0);

	form_status = XtVaCreateWidget("form_status",
		xmFormWidgetClass,	form,
                XmNtopAttachment,       XmATTACH_FORM,
                XmNtopOffset,		1,
		XmNrightAttachment,	XmATTACH_FORM,
                XmNleftAttachment,      XmATTACH_FORM,
                XmNfractionBase,	1,
		XmNmarginHeight,	0,
                NULL);

	line_label = XtVaCreateManagedWidget("line_label",
                xmLabelWidgetClass,     form_status,
                XmNalignment,           XmALIGNMENT_BEGINNING,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_FORM,
		XmNleftOffset,		10,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
		XmNmarginHeight,	1,
		NULL);

	line_text = XtVaCreateManagedWidget("line_num",
		xmTextWidgetClass,	form_status,
		XmNeditMode,		XmSINGLE_LINE_EDIT,
		XmNshadowThickness,	0,
		XmNcolumns,		4,
		XmNresizeWidth,		True,
		XmNeditable,		False,
		XmNcursorPositionVisible,	False,
		XmNautoShowCursorPosition,	False,
		XmNmarginHeight,	1,
		XmNtraversalOn,		False,
		XmNselectionArray,	selectionArray,
		XmNselectionArrayCount, 2,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_WIDGET,
		XmNleftWidget,		line_label,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
		NULL);

#if (XmVersion >= 1002)	/* Motif 1.2 */
	XtSetArg(args[0], XmNdragInitiatorProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[1], XmNdragReceiverProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[2], XmNdropSiteActivity, XmDROP_SITE_INACTIVE);

	XmDropSiteUpdate(line_text, args, 3);
#endif

	col_label = XtVaCreateManagedWidget("col_label",
                xmLabelWidgetClass,     form_status,
                XmNalignment,           XmALIGNMENT_BEGINNING,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_WIDGET,
		XmNleftWidget,		line_text,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
		XmNmarginHeight,	1,
		NULL);

	col_text = XtVaCreateManagedWidget("col_num",
		xmTextWidgetClass,	form_status,
		XmNeditMode,		XmSINGLE_LINE_EDIT,
		XmNshadowThickness,	0,
		XmNcolumns,		3,
		XmNresizeWidth,		True,
		XmNeditable,		False,
		XmNcursorPositionVisible,	False,
		XmNautoShowCursorPosition,	False,
		XmNmarginHeight,	1,
		XmNtraversalOn,		False,
		XmNselectionArray,	selectionArray,
		XmNselectionArrayCount, 2,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_WIDGET,
		XmNleftWidget,		col_label,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
		NULL);

#if (XmVersion >= 1002)	/* Motif 1.2 */
	XtSetArg(args[0], XmNdragInitiatorProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[1], XmNdragReceiverProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[2], XmNdropSiteActivity, XmDROP_SITE_INACTIVE);

	XmDropSiteUpdate(col_text, args, 3);
#endif

	env_line_1 = XtVaCreateManagedWidget("env_line_1",
		xmTextWidgetClass,	form_status,
		XmNeditMode,		XmSINGLE_LINE_EDIT,
		XmNshadowThickness,	1,
		XmNcolumns,		11,
		XmNresizeWidth,		True,
		XmNeditable,		False,
		XmNcursorPositionVisible,	False,
		XmNautoShowCursorPosition,	False,
		XmNmarginHeight,	1,
		XmNtraversalOn,		False,
		XmNselectionArray,	selectionArray,
		XmNselectionArrayCount, 2,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_WIDGET,
		XmNleftWidget,		col_text,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
                NULL);

#if (XmVersion >= 1002)	/* Motif 1.2 */
	XtSetArg(args[0], XmNdragInitiatorProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[1], XmNdragReceiverProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[2], XmNdropSiteActivity, XmDROP_SITE_INACTIVE);

	XmDropSiteUpdate(env_line_1, args, 3);
#endif

	env_line_2 = XtVaCreateManagedWidget("env_line_2",
		xmTextWidgetClass,	form_status,
		XmNeditMode,		XmSINGLE_LINE_EDIT,
		XmNshadowThickness,	1,
		XmNcolumns,		12,
		XmNresizeWidth,		True,
		XmNeditable,		False,
		XmNcursorPositionVisible,	False,
		XmNautoShowCursorPosition,	False,
		XmNmarginHeight,	1,
		XmNtraversalOn,		False,
		XmNselectionArray,	selectionArray,
		XmNselectionArrayCount, 2,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_WIDGET,
		XmNleftWidget,		env_line_1,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
                NULL);

#if (XmVersion >= 1002)	/* Motif 1.2 */
	XtSetArg(args[0], XmNdragInitiatorProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[1], XmNdragReceiverProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[2], XmNdropSiteActivity, XmDROP_SITE_INACTIVE);

	XmDropSiteUpdate(env_line_1, args, 3);
#endif

	status_line = XtVaCreateManagedWidget("status_line",
		xmTextWidgetClass,	form_status,
		XmNeditMode,		XmSINGLE_LINE_EDIT,
		XmNshadowThickness,	1,
		XmNeditable,		False,
		XmNcursorPositionVisible,	False,
		XmNautoShowCursorPosition,	False,
		XmNmarginHeight,	1,
		XmNtraversalOn,		False,
		XmNselectionArray,	selectionArray,
		XmNselectionArrayCount, 2,
		XmNtopAttachment,	XmATTACH_POSITION,
		XmNtopPosition,		0,
		XmNleftAttachment,	XmATTACH_WIDGET,
		XmNleftWidget,		env_line_2,
		XmNrightAttachment,	XmATTACH_FORM,
		XmNrightOffset,		2,
		XmNbottomAttachment,	XmATTACH_POSITION,
		XmNbottomPosition,	1,
                NULL);

#if (XmVersion >= 1002)	/* Motif 1.2 */
	XtSetArg(args[0], XmNdragInitiatorProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[1], XmNdragReceiverProtocolStyle, XmDRAG_NONE);
	XtSetArg(args[2], XmNdropSiteActivity, XmDROP_SITE_INACTIVE);

	XmDropSiteUpdate(status_line, args, 3);
#endif

	XtManageChild(form_status);

        separator = XtVaCreateManagedWidget("Separator",
                xmSeparatorGadgetClass, form,
                XmNtopAttachment,       XmATTACH_WIDGET,
                XmNtopWidget,           form_status,
                XmNtopOffset,		1,
                XmNleftAttachment,      XmATTACH_FORM,
                XmNrightAttachment,     XmATTACH_FORM,
                NULL);

	XtSetArg(args[0], XmNeditMode,		XmMULTI_LINE_EDIT);
	XtSetArg(args[1], XmNrows,		25);
	XtSetArg(args[2], XmNcolumns,		79);
	XtSetArg(args[3], XmNtopAttachment,     XmATTACH_WIDGET);
	XtSetArg(args[4], XmNtopWidget,         separator);
	XtSetArg(args[5], XmNtopOffset,         1);
	XtSetArg(args[6], XmNleftAttachment,    XmATTACH_FORM);
	XtSetArg(args[7], XmNrightAttachment,   XmATTACH_FORM);
        XtSetArg(args[8], XmNbottomAttachment,  XmATTACH_FORM);

	text_w = XmCreateScrolledText(form, "text_w", args, 9);

	XtAppAddActions(app_context, text_actions, XtNumber(text_actions));

	XtAddCallback(text_w, XmNmodifyVerifyCallback, Modify_cb, (XtPointer)NULL);
	XtAddCallback(text_w, XmNmotionVerifyCallback, Motion_cb, (XtPointer)NULL);

	XtSetArg(args[0], XmNspacing,	0);
	XtSetValues(XtParent(text_w), args, 1);

	XtSetArg(args[0], XmNhorizontalScrollBar, &ScrollBar);
	XtGetValues(XtParent(text_w), args, 1);

	XtSetArg(args[0], XmNshadowThickness,	1);
	XtSetValues(ScrollBar, args, 1);

	XtSetArg(args[0], XmNverticalScrollBar, &ScrollBar);
	XtGetValues(XtParent(text_w), args, 1);

	XtSetArg(args[0], XmNshadowThickness,	1);
	XtSetValues(ScrollBar, args, 1);

	/*
	 *	Create Popup Menu
	 */

	popup_menu = XmVaCreateSimplePopupMenu(text_w,
		"popup_menu", PopupMenu_cb,
		XmVaTITLE,	NULL,
		XmVaSEPARATOR,
		XmVaSEPARATOR,
		XmVaCASCADEBUTTON, NULL, NULL,		/* Undo	Menu	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Redo		*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Cut 		*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Copy 	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Paste 	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL, /* Clear 	*/
		XmNpopupEnabled,	False,
		NULL);

	popup_undo_button = XtNameToWidget(popup_menu, "button_0");
	popup_redo_button = XtNameToWidget(popup_menu, "button_1");

	popup_submenu = XmVaCreateSimplePulldownMenu(popup_menu,
		"popup_submenu", 0, PopupSubMenu_cb,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Undo Last	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Undo All	*/
		NULL);

	XtVaSetValues(XtNameToWidget(popup_menu, "button_0"),
		XmNsubMenuId, popup_submenu,
		NULL);

	XtAddEventHandler(text_w, ButtonPressMask, False,
			(XtEventHandler)popup_cb, (XtPointer)popup_menu);
	/*
	 *	Manage TextWidget
	 */

	XtManageChild(text_w);

	/*
	 *	File Open DialogBox
	 */

	XtSetArg(args[0], XmNdialogStyle,       XmDIALOG_FULL_APPLICATION_MODAL);
        XtSetArg(args[1], XmNautoUnmanage,	False);
	XtSetArg(args[2], XmNnoResize,		True);
	XtSetArg(args[3], XmNresizePolicy,      XmRESIZE_NONE);

	OpenDialog = XmCreateFileSelectionDialog(text_w,
		"OpenDialog", args, 4);

	XtUnmanageChild(XmFileSelectionBoxGetChild(OpenDialog,
			XmDIALOG_HELP_BUTTON));
	XtAddCallback(OpenDialog, XmNokCallback, open_cb, (XtPointer)NULL);
	XtAddCallback(OpenDialog, XmNcancelCallback,
			CloseBox, (XtPointer)NULL);

	/*
	 *	Save As... DialogBox
	 */

	XtSetArg(args[0], XmNdialogStyle,       XmDIALOG_FULL_APPLICATION_MODAL);
        XtSetArg(args[1], XmNautoUnmanage,	False);
	XtSetArg(args[2], XmNnoResize,		True);
	XtSetArg(args[3], XmNresizePolicy,      XmRESIZE_NONE);

	SaveAsDialog = XmCreateFileSelectionDialog(text_w,
		"SaveAsDialog", args, 4);

	XtUnmanageChild(XmFileSelectionBoxGetChild(SaveAsDialog,
			XmDIALOG_HELP_BUTTON));
	XtAddCallback(SaveAsDialog, XmNokCallback, SaveAs_cb,
			(XtPointer)NULL);
	XtAddCallback(SaveAsDialog, XmNapplyCallback, OpenDir_cb,
			(XtPointer)NULL);
	XtAddCallback(SaveAsDialog, XmNcancelCallback,
			CloseBox, (XtPointer)NULL);

	/*
	 *	Overwrite Question DialogBox
	 */

	XtSetArg(args[0], XmNdialogStyle, 	XmDIALOG_FULL_APPLICATION_MODAL);
        XtSetArg(args[1], XmNautoUnmanage,	False);
	XtSetArg(args[2], XmNnoResize,		True);
	XtSetArg(args[3], XmNresizePolicy,	XmRESIZE_NONE);

	OverwriteDialog = XmCreateQuestionDialog(text_w, "overwriteQuestionDialog",
			args, 4);

	XtUnmanageChild(XmMessageBoxGetChild(OverwriteDialog,
			XmDIALOG_HELP_BUTTON));
	XtAddCallback(OverwriteDialog, XmNokCallback, OverWrite_cb,
			(XtPointer)NULL);
	XtAddCallback(OverwriteDialog, XmNcancelCallback,
			CloseBox, (XtPointer)NULL);

	/*
	 *	Question DialogBox
	 */

	XtSetArg(args[0], XmNdialogStyle, 	XmDIALOG_FULL_APPLICATION_MODAL);
        XtSetArg(args[1], XmNautoUnmanage,	False);
	XtSetArg(args[2], XmNnoResize,		True);
	XtSetArg(args[3], XmNresizePolicy,	XmRESIZE_NONE);

	QuestionDialog = XmCreateQuestionDialog(text_w, "QuestionDialog",
			args, 4);

	XtAddCallback(QuestionDialog, XmNokCallback, yes_cb, (XtPointer)NULL);
	XtAddCallback(QuestionDialog, XmNcancelCallback, no_cb, (XtPointer)NULL);
	XtAddCallback(QuestionDialog, XmNhelpCallback,
			CloseBox, (XtPointer)NULL);

	/*
	 *	Error Dialog
	 */

	XtSetArg(args[0], XmNdialogStyle, 	XmDIALOG_FULL_APPLICATION_MODAL);
	XtSetArg(args[1], XmNnoResize,		True);
	XtSetArg(args[2], XmNresizePolicy,	XmRESIZE_NONE);

	ErrorDialog = XmCreateErrorDialog(text_w, "ErrorDialog",
		args, 3);

	XtUnmanageChild(XmMessageBoxGetChild(ErrorDialog,
			XmDIALOG_CANCEL_BUTTON));
	XtUnmanageChild(XmMessageBoxGetChild(ErrorDialog,
			XmDIALOG_HELP_BUTTON));

	/*
	 *	Go Dialog
	 */

	XtSetArg(args[0], XmNautoUnmanage,	True);
	XtSetArg(args[1], XmNnoResize,		True);
	XtSetArg(args[2], XmNresizePolicy,	XmRESIZE_NONE);

	go_dialog = XmCreatePromptDialog(text_w, "GoDialog",
			args, 3);

	XtUnmanageChild(XmSelectionBoxGetChild(go_dialog,
			XmDIALOG_HELP_BUTTON));
	XtAddCallback(go_dialog, XmNokCallback, go_cb, (XtPointer)NULL);
	XtAddCallback(go_dialog, XmNcancelCallback,
			CloseSelectionBox, (XtPointer)NULL);
        XtAddCallback(XmSelectionBoxGetChild(go_dialog, XmDIALOG_TEXT),
        	XmNmodifyVerifyCallback, ValueChange_cb,
                (XtPointer)XmSelectionBoxGetChild(go_dialog, XmDIALOG_OK_BUTTON));
        XtSetSensitive(XmSelectionBoxGetChild(go_dialog, XmDIALOG_OK_BUTTON),
                False);

	/*
	 *	Search  & Replace Dialog
	 */

	SearchDialog = smCreateSearchDialog(text_w, "SearchDialog");

	/*
	 *	TabSize Dialog
	 */
	 buf = XtMalloc(BUFSIZ);
	(void)sprintf(buf, "%d", TabSize);
	str = XmStringCreateSimple(buf);
	XtFree(buf);

	XtSetArg(args[0], XmNdialogStyle, 	XmDIALOG_FULL_APPLICATION_MODAL);
	XtSetArg(args[1], XmNnoResize,		True);
	XtSetArg(args[2], XmNresizePolicy,	XmRESIZE_NONE);
	XtSetArg(args[3], XmNtextString,	str);

	tab_dialog = XmCreatePromptDialog(text_w, "TabSizeDialog",
			args, 4);

	XtUnmanageChild(XmSelectionBoxGetChild(tab_dialog,
			XmDIALOG_HELP_BUTTON));
	XtAddCallback(tab_dialog, XmNokCallback, settab_cb, (XtPointer)NULL);
	XtAddCallback(tab_dialog, XmNcancelCallback,
			CloseSelectionBox, (XtPointer)NULL);
        XtAddCallback(XmSelectionBoxGetChild(tab_dialog, XmDIALOG_TEXT),
                XmNmodifyVerifyCallback, ValueChange_cb,
                (XtPointer)XmSelectionBoxGetChild(tab_dialog,
                        XmDIALOG_OK_BUTTON));

	XmStringFree(str);

	/*
	 *	Block Indent Dialog
	 */
	 buf = XtMalloc(BUFSIZ);
	(void)sprintf(buf, "%d", BlockIndentSize);
        str = XmStringCreateSimple(buf);
	XtFree(buf);

	XtSetArg(args[0], XmNdialogStyle, 	XmDIALOG_FULL_APPLICATION_MODAL);
	XtSetArg(args[1], XmNnoResize,		True);
	XtSetArg(args[2], XmNresizePolicy,	XmRESIZE_NONE);
	XtSetArg(args[3], XmNtextString,	str);

	block_dialog = XmCreatePromptDialog(text_w, "BlockDialog",
			args, 4);

	XtUnmanageChild(XmSelectionBoxGetChild(block_dialog,
			XmDIALOG_HELP_BUTTON));
	XtAddCallback(block_dialog, XmNokCallback, setblock_cb, (XtPointer)NULL);
	XtAddCallback(block_dialog, XmNcancelCallback,
			CloseSelectionBox, (XtPointer)NULL);
        XtAddCallback(XmSelectionBoxGetChild(block_dialog, XmDIALOG_TEXT),
                XmNmodifyVerifyCallback, ValueChange_cb,
                (XtPointer)XmSelectionBoxGetChild(block_dialog,
                        XmDIALOG_OK_BUTTON));

	XmStringFree(str);

	/*
	 *	Help Dialog
	 */

	help_dialog = CreateHelpDialog(text_w, "HelpDialog");

	/*
	 *	About Dialog
	 */

	str = XmStringCreateLtoR(ABOUT_STR, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(args[0], XmNmessageString,	str);
	XtSetArg(args[1], XmNdialogStyle, 	XmDIALOG_FULL_APPLICATION_MODAL);
	XtSetArg(args[2], XmNnoResize,		True);
	XtSetArg(args[3], XmNresizePolicy,	XmRESIZE_NONE);
	XtSetArg(args[4], XmNmessageAlignment,	XmALIGNMENT_CENTER);

	about_dialog = XmCreateInformationDialog(text_w, "AboutDialog",
			args, 5);

	XmStringFree(str);

	XtUnmanageChild(XmMessageBoxGetChild(about_dialog,
			XmDIALOG_CANCEL_BUTTON));
	XtUnmanageChild(XmMessageBoxGetChild(about_dialog,
			XmDIALOG_HELP_BUTTON));

	FileName = XtNewString("(NONE)");

	update_StatusLine(text_w, 0);

        line = -1;
        is_Empty = True;
        for (i = 1; i < argc; i++) {
            buf = XtNewString(argv[i]);
            if (buf[0] == '+') {
                if (!strcmp(buf + 1, "insert"))		/* Insert Mode Off */
                    insert_mode = False;
                else if (!strcmp(buf + 1, "indent"))	/* Indent Mode Off */
                    indent_mode = False;
                else if (!strcmp(buf + 1, "tab"))	/* Use tab character Mode Off */
                    fill_mode = False;
                else if (!strcmp(buf + 1, "undo"))	/* Undo after Save Off */
                    undo_after_save = False;
                else if (!strcmp(buf + 1, "showmatch"))	/* Show Match Off */
                    show_match = False;
                else if (!strcmp(buf + 1, "backup"))	/* Create Bkacup Off */
                    create_backup = False;
                else if (!strcmp(buf + 1, "rw"))	/* Remove Whitespace Off */
                    optimize_text = False;
                else if (is_num(buf + 1) == True)
                    line = atoi(buf + 1);
                else usage(argv[0]);
            }
            else if (buf[0] == '-') {
                if (!strcmp(buf + 1, "insert"))		/* Insert Mode On */
                    insert_mode = True;
                else if (!strcmp(buf + 1, "indent"))	/* Indent Mode On */
                    indent_mode = True;
                else if (!strcmp(buf + 1, "tab"))	/* Use tab character Mode On */
                    fill_mode = True;
                else if (!strcmp(buf + 1, "undo"))	/* Undo after Save On */
                    undo_after_save = True;
                else if (!strcmp(buf + 1, "showmatch"))	/* Show Match On */
                    show_match = True;
                else if (!strcmp(buf + 1, "backup"))	/* Create Bkacup On */
                    create_backup = True;
                else if (!strcmp(buf + 1, "rw"))	/* Remove Whitespace On */
                    optimize_text = True;
                else if (!strcmp(buf + 1, "tabs") || !strcmp(buf + 1, "ts")) {
			/* Set Tab size */
                    i++;
                    TabSize = atoi(argv[i]);
	            if (TabSize > MAXTAB || TabSize <= 0) {
	                TabSize = 8;
	                printf("Tab Size : Invalid Tab size\n");
	            }
                }
                else if (!strcmp(buf + 1, "blocki") || !strcmp(buf + 1, "bi")) {
			/* Set Block indent */
                    i++;
                    BlockIndentSize = atoi(argv[i]);
	            if (BlockIndentSize > MAXTAB || BlockIndentSize <= 0) {
	                BlockIndentSize = 8;
	                printf("Block indent : Invalid size\n");
	            }
                }
                else usage(argv[0]);	/* Help */
	    }
	    else {
                if (is_Empty) {
                    is_Empty = False;
		    ignoreModify = True;
                    XtFree(FileName);
                    FileName = XtNewString(buf);
                    if (!open_file(FileName))
                        XmTextSetString(status_line, "New File");
                    add_name(FileName);
		    ignoreModify = False;
                }
            }
            XtFree(buf);
        }

	SetTabSize(text_w, TabSize);
	ignoreModify = True;
	XmTextInsert(text_w, 0, "\n");
	XmTextReplace(text_w, 0, 1, "");
	ignoreModify = False;

        if (line > 0)
            smTextSetInsertionLine(text_w, line);

	/* Environment */

  	XmToggleButtonGadgetSetState(XtNameToWidget(option_env, "button_0"),
  		insert_mode, True);
  	XmToggleButtonGadgetSetState(XtNameToWidget(option_env, "button_1"),
  		indent_mode, True);
  	XmToggleButtonGadgetSetState(XtNameToWidget(option_env, "button_2"),
  		fill_mode, True);
  	XmToggleButtonGadgetSetState(XtNameToWidget(option_env, "button_3"),
  		undo_after_save, False);
  	XmToggleButtonGadgetSetState(XtNameToWidget(option_env, "button_4"),
  		show_match, False);

	/* Preference */

	XmToggleButtonGadgetSetState(XtNameToWidget(option_pref, "button_0"),
		create_backup, False);

	XmToggleButtonGadgetSetState(XtNameToWidget(option_pref, "button_1"),
		optimize_text, False);

	update_EnvironmentLine();

	XtSetSensitive(undo_button, False);
	XtSetSensitive(redo_button, False);
	XtSetSensitive(popup_undo_button, False);
	XtSetSensitive(popup_redo_button, False);

	XtManageChild(form);
	XtManageChild(main_w);
	XtRealizeWidget(toplevel);

	set_title(FileName);

	XtAppMainLoop(app_context);
}

#ifdef _NO_PROTO
Widget CreateMenuBar(parent)
Widget parent;
#else
Widget CreateMenuBar(Widget parent)
#endif
{
	Widget menu;

	menu = XmVaCreateSimpleMenuBar(parent, "MainMenu",
		XmVaCASCADEBUTTON, NULL, NULL,		/* File 	*/
		XmVaCASCADEBUTTON, NULL, NULL,		/* Edit 	*/
		XmVaCASCADEBUTTON, NULL, NULL,		/* Search 	*/
		XmVaCASCADEBUTTON, NULL, NULL,		/* Options 	*/
		XmVaCASCADEBUTTON, NULL, NULL,		/* Help 	*/
		NULL);

	file_menu = XmVaCreateSimplePulldownMenu(menu,
		"file_menu", 0, FileMenu_cb,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* New		*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Open		*/
		XmVaCASCADEBUTTON, NULL, NULL,		/* List... 	*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Save		*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Save As.. 	*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Quit 	*/
		NULL);

	pick_submenu = XmCreatePulldownMenu(file_menu,
		"pick_submenu", (ArgList)NULL, 0);

	XtVaSetValues(XtNameToWidget(file_menu, "button_2"),
		XmNsubMenuId, pick_submenu,
		NULL);

	XtSetSensitive(XtNameToWidget(file_menu, "button_2"), False);

	edit_menu = XmVaCreateSimplePulldownMenu(menu,
		"edit_menu", 1, EditMenu_cb,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Undo 	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Redo 	*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Cut 		*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Copy 	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Paste 	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL, /* Clear 	*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Select All	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Unselect 	*/
		NULL);

	undo_button = XtNameToWidget(edit_menu, "button_0");
	redo_button = XtNameToWidget(edit_menu, "button_1");

	search_menu = XmVaCreateSimplePulldownMenu(menu,
		"search_menu", 2, SearchMenu_cb,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Find		*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Replace	*/
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Find again	*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Go to line	*/
		NULL);

	option_menu = XmVaCreateSimplePulldownMenu(menu,
		"option_menu", 3, OptionMenu_cb,
		XmVaCASCADEBUTTON, NULL, NULL,		/* Environment */
		XmVaCASCADEBUTTON, NULL, NULL,		/* Preference */
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Save setting	*/
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Refresh	*/
		NULL);

	option_env = XmVaCreateSimplePulldownMenu(option_menu,
		"option_env", 0, EnvMenu_cb,
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Insert mode */
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Indent mode */
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Use tab character */
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Undo after save */
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Show match */
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Pending delete */
		XmVaSEPARATOR,
		XmVaPUSHBUTTON,  NULL, NULL, NULL, NULL, /* Tab size */
		XmVaPUSHBUTTON,  NULL, NULL, NULL, NULL, /* Block Indent size */
		NULL);

	XtVaSetValues(XtNameToWidget(option_env, "button_0"),
		XmNset,	insert_mode,
		NULL);

	XtVaSetValues(XtNameToWidget(option_env, "button_1"),
		XmNset,	indent_mode,
		NULL);

	XtVaSetValues(XtNameToWidget(option_env, "button_2"),
		XmNset,	fill_mode,
		NULL);

	XtVaSetValues(XtNameToWidget(option_env, "button_3"),
		XmNset,	undo_after_save,
		NULL);

	XtVaSetValues(XtNameToWidget(option_env, "button_4"),
		XmNset,	show_match,
		NULL);

	XtVaSetValues(XtNameToWidget(option_env, "button_5"),
		XmNset,	pending_delete,
		NULL);

	XtVaSetValues(XtNameToWidget(option_menu, "button_0"),
		XmNsubMenuId,	option_env,
		NULL);

	option_pref = XmVaCreateSimplePulldownMenu(option_menu,
		"option_pref", 1, PrefMenu_cb,
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Create backup file */
		XmVaCHECKBUTTON, NULL, NULL, NULL, NULL, /* Remove whitespace */
		NULL);

	XtVaSetValues(XtNameToWidget(option_pref, "button_0"),
		XmNset,	create_backup,
		NULL);

	XtVaSetValues(XtNameToWidget(option_pref, "button_1"),
		XmNset,	optimize_text,
		NULL);

	XtVaSetValues(XtNameToWidget(option_menu, "button_1"),
		XmNsubMenuId,	option_pref,
		NULL);

	help_menu = XmVaCreateSimplePulldownMenu(menu,
		"help_menu", 4, HelpMenu_cb,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* Contents... */
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, NULL, NULL, NULL, NULL,	/* About... */
		NULL);

	XtVaSetValues(menu,
		XmNmenuHelpWidget, XtNameToWidget(menu, "button_4"),
		NULL);

        return (menu);
}

/*
 *	XtEnevtHandler
 */
#ifdef _NO_PROTO
void popup_cb(w, client_data, event, continue_to_dispatch_return)
Widget w;
XtPointer client_data;
XButtonPressedEvent *event;
Boolean *continue_to_dispatch_return;
#else
void popup_cb(Widget w, XtPointer client_data,
		XButtonPressedEvent *event,
		Boolean *continue_to_dispatch_return)
#endif
{
	Widget popup = (Widget)client_data;
        Boolean is_selected;
        XmTextPosition start, end;

	if (event->button == 3) {
            is_selected = XmTextGetSelectionPosition(w, &start, &end);
            if (is_selected && start != end) {
                XtSetSensitive(XtNameToWidget(popup_menu, "button_2"), True);
                XtSetSensitive(XtNameToWidget(popup_menu, "button_3"), True);
                XtSetSensitive(XtNameToWidget(popup_menu, "button_5"), True);
            }
            else {
                XtSetSensitive(XtNameToWidget(popup_menu, "button_2"), False);
                XtSetSensitive(XtNameToWidget(popup_menu, "button_3"), False);
                XtSetSensitive(XtNameToWidget(popup_menu, "button_5"), False);
            }
	    XmMenuPosition(popup, event);
	    XtManageChild(popup);
	}
}

#ifdef _NO_PROTO
void usage(name)
char *name;
#else
void usage(char *name)
#endif
{
        printf("usage: %s [+line] [-/+insert] [-/+indent ] [-/+tab] [-/+undo] [-/+showmatch]\n", name);
        printf("\t[-tabs n] [-blocki n] [-/+backup] [-/+rw] [filename]\n");
        exit(0);
}

void update_EnvironmentLine()
{
        char buf[BUFSIZ];

        if (!file_saved) sprintf(buf, "* ");
	else sprintf(buf, "  ");

	if (insert_mode) strcat(buf, "Insert");
	else strcat(buf, "Overwrite");

	XmTextSetString(env_line_1, buf);

	if (indent_mode) sprintf(buf, "Indent  ");
	else sprintf(buf, "Unindent");

        if (fill_mode) strcat(buf, " Tab");

	XmTextSetString(env_line_2, buf);
}

#ifdef _NO_PROTO
void update_StatusLine(w, pos)
Widget w;
XmTextPosition pos;
#else
void update_StatusLine(Widget w, XmTextPosition pos)
#endif
{
        int line, col;
	char buf[10];

        smPosToLineColumn(w, pos, TabSize, &line, &col);

	(void)sprintf(buf, "%d", line);
	XmTextSetString(line_text, buf);

	(void)sprintf(buf, "%d", col);
	XmTextSetString(col_text, buf);
}

#ifdef _NO_PROTO
void add_name(name)
char *name;
#else
void add_name(char *name)
#endif
{
	char buf[BUFSIZ], *str;
	XmString label_str;
	Boolean match;
	static Boolean full = False;
        int i, ret, max_list;

	if (XtIsSensitive(XtNameToWidget(file_menu, "button_2")) == False)
	    XtSetSensitive(XtNameToWidget(file_menu, "button_2"), True);

	max_list = full ? MAXLIST : file_count;
	match = True;
	for (i = 0; i < max_list; i++) {
	    XtVaGetValues(pick_button[i],
	    	XmNlabelString,	        &label_str,
	    	NULL);
	    XmStringGetLtoR(label_str, XmSTRING_DEFAULT_CHARSET, &str);
	    ret = strcmp(str + 3, name);
	    XmStringFree(label_str);
	    XtFree(str);
	    if (!ret) {
	        match = False;
	        break;
	    }
	}

	if (match) {
	    str = XtMalloc(strlen(name) + 4);
	    sprintf(str, "%d. %s", file_count, name);
	    label_str = XmStringCreateSimple(str);
	    if (!full) {
	        (void)sprintf(buf, "button_%d", file_count);
	        pick_button[file_count] = XtVaCreateManagedWidget(buf,
			xmPushButtonGadgetClass, pick_submenu,
		        XmNlabelString, label_str,
		        NULL);
	        XtAddCallback(pick_button[file_count],
			XmNactivateCallback, pick_cb, (XtPointer)NULL);
                XtVaSetValues(pick_button[file_count],
            	    XmNlabelString,	        label_str,
		    XmNmnemonic,		str[0],
		    NULL);
	    }
	    else {
	        XtVaSetValues(pick_button[file_count],
		    XmNlabelString,	        label_str,
		    XmNmnemonic,		str[0],
		    NULL);
            }
	    XmStringFree(label_str);
	    XtFree(str);
	    if (file_count == (MAXLIST - 1))
	        full = True;
	    file_count = (file_count + 1 + MAXLIST) % MAXLIST;
	}
}

#ifdef _NO_PROTO
void update_SaveAsDialog_data(w, str)
Widget w;
char *str;
#else
void update_SaveAsDialog_data(Widget w, char *str)
#endif
{
	XmString dirSpec, dirPath, dirMask, dirFile;
        char *f_name, *f_path;

        if (strcmp(str, "(NONE)")) {
            parseFilename(str, &f_path, &f_name);

	    if (!strcmp(f_path, ""))
	        dirMask = XmStringCreateSimple("*");
	    else dirMask = XmStringCreateSimple(f_path);

	    if (!strcmp(f_name, ""))
	        dirFile = XmStringCreateSimple(str);
	    else dirFile = XmStringCreateSimple(f_name);

	    XtFree(f_path);
	    XtFree(f_name);
        }
        else {
            dirMask = XmStringCreateSimple("*");
            dirFile = XmStringCreateSimple("UNNAMED");
        }
	XmFileSelectionDoSearch(w, dirMask);
	XmStringFree(dirMask);

	XtVaGetValues(w,
		XmNdirectory, &dirPath,
		NULL);

        dirSpec = XmStringConcat(dirPath, dirFile);
	XmStringFree(dirPath);
	XmStringFree(dirFile);

	XtVaSetValues(w,
		XmNdirSpec,	dirSpec,
		NULL);
	XmStringFree(dirSpec);
}

#ifdef _NO_PROTO
void set_QuestionDialog_message(w, str)
Widget w;
char *str;
#else
void set_QuestionDialog_message(Widget w, char *str)
#endif
{
	char *buf;

	buf = XtMalloc((Cardinal)strlen(str) + 50);
	(void)sprintf(buf, "File:%s\n\nFile has been modfied.\nSave it now ?", str);
	setDialogMessage(w, buf);
	XtFree(buf);
}

#ifdef _NO_PROTO
void set_ErrorDialog_message(w, str, tag)
Widget w;
char *str;
int tag;
#else
void set_ErrorDialog_message(Widget w, char *str, int tag)
#endif
{
	char *buf;

	buf = XtMalloc((Cardinal)strlen(str) + strlen(sys_errlist[errno]) + 40);
	switch (tag) {
	    case FILE_SAVE :
		(void)sprintf(buf, "File:%s\n\nUnable to save file.\n%s", str, sys_errlist[errno]);
		break;
	    case FILE_OPEN :
		(void)sprintf(buf, "File:%s\n\nUnable to open file.\n%s", str, sys_errlist[errno]);
		break;
	}
	setDialogMessage(w, buf);
	XtFree(buf);
}

#ifdef _NO_PROTO
void set_OverwriteDialog_message(w, str)
Widget w;
char *str;
#else
void set_OverwriteDialog_message(Widget w, char *str)
#endif
{
	char *buf;

	buf = XtMalloc((Cardinal)strlen(OpenName) + 60);
	(void)sprintf(buf, "File:%s\n\nThat file exits and has data in it.\nReplace it ?", str);
	setDialogMessage(w, buf);
	XtFree(buf);
}

#ifdef _NO_PROTO
void set_title(name)
char *name;
#else
void set_title(char *name)
#endif
{
	char *buf;

	buf = XtMalloc((Cardinal)strlen(name) + 15);
	(void)sprintf(buf, "SM/Editor - %s", name);
	XStoreName(XtDisplay(toplevel), XtWindow(toplevel), buf);
	XSetIconName(XtDisplay(toplevel), XtWindow(toplevel), buf);
	XtFree(buf);
}

#ifdef _NO_PROTO
void UndoLast(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void UndoLast(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition pos;

	if (undo) {
	    ignoreModify = True;
	    XmTextSetEditable(w, False);

	    if (found) {
	        XmTextClearSelection(w, event->xkey.time);
	        found = False;
	    }
	    XtSetSensitive(redo_button, True);
	    XtSetSensitive(popup_redo_button, True);
            XmTextDisableRedisplay(w);

	    redo = NewLink(redo, undo->pos);
	    if (undo->length > 0L) {
	        XmTextReplace(w, (XmTextPosition)undo->pos,
	    		undo->pos + undo->length, "");
	    }
	    else {
	        XmTextInsert(w, (XmTextPosition)undo->pos, undo->buffer);
	        redo->length = strlen(undo->buffer);
	    }
	    redo->buffer = XtNewString(undo->buffer);
	    XtFree(undo->buffer);
	    pos = (XmTextPosition)undo->pos;
	    undo = DeleteCurrentLink(undo);
	    if (undo && undo->length == -1) {
	        redo = NewLink(redo, undo->pos);
	        XmTextInsert(w, (XmTextPosition)undo->pos, undo->buffer);
	        redo->length = -1;
                redo->buffer = XtNewString(undo->buffer);
                XtFree(undo->buffer);
                pos = (XmTextPosition)undo->pos;
                undo = DeleteCurrentLink(undo);
	    }
	    if (!undo) {
	        XtSetSensitive(undo_button, False);
	        XtSetSensitive(popup_undo_button, False);
	    }
            XmTextEnableRedisplay(w);
	    XmTextSetInsertionPosition(w, pos);

	    XmTextSetEditable(w, True);
	    ignoreModify = False;
            if (undo_after_save && file_saved) {
                file_saved = False;
                update_EnvironmentLine();
            }
        }
	XFlush(event->xkey.display);

	pos = XmTextGetInsertionPosition(w);
	update_StatusLine(w, pos);
}

#ifdef _NO_PROTO
void UndoAll(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void UndoAll(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	static XmTextPosition position;
	XmTextPosition pos;

	if (undo) {
	    ignoreModify = True;
	    if (found) {
	        XmTextClearSelection(w, event->xkey.time);
	        found = False;
	    }
	    XtSetSensitive(redo_button, True);
	    XtSetSensitive(popup_redo_button, True);
	    while (undo) {
	        redo = NewLink(redo, undo->pos);
	        if (undo->length > 0L) {
	            XmTextReplace(w, (XmTextPosition)undo->pos,
			(XmTextPosition)undo->pos + undo->length, "");
	        }
	        else if (undo->length == -1) {
	            XmTextInsert(w, (XmTextPosition)undo->pos, undo->buffer);
	            redo->length = -1;
	        }
	        else {
	            XmTextInsert(w, (XmTextPosition)undo->pos, undo->buffer);
	            redo->length = strlen(undo->buffer);
	        }
                redo->buffer = XtNewString(undo->buffer);
                XtFree(undo->buffer);
	        position = (XmTextPosition)undo->pos;
                undo = DeleteCurrentLink(undo);
	    }
	    XtSetSensitive(undo_button, False);
	    XtSetSensitive(popup_undo_button, False);

	    XmTextSetInsertionPosition(w, position);
	    ignoreModify = False;
            if (undo_after_save && file_saved) {
                file_saved = False;
                update_EnvironmentLine();
            }
	}
	XFlush(event->xkey.display);

	pos = XmTextGetInsertionPosition(w);
	update_StatusLine(w, pos);
}

#ifdef _NO_PROTO
void RedoAction(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void RedoAction(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition pos;

	if (redo) {
	    ignoreModify = True;
	    if (found) {
	        XmTextClearSelection(w, CurrentTime);
	        found = False;
	    }
	    XtSetSensitive(undo_button, True);
	    XtSetSensitive(popup_undo_button, True);
            XmTextDisableRedisplay(w);

    	    undo = NewLink(undo, (long)redo->pos);
	    if (redo->length > 0L) {
	        XmTextReplace(w, (XmTextPosition)redo->pos,
			(XmTextPosition)redo->pos + redo->length, "");
	    }
	    else {
    	        if (redo->length == -1) {
    	             XmTextReplace(w, (XmTextPosition)redo->pos,
			(XmTextPosition)redo->pos + strlen(redo->buffer), "");
    	            undo->length = -1;
                    undo->buffer = XtNewString(redo->buffer);
                    XtFree(redo->buffer);
                    redo = DeleteCurrentLink(redo);
    	            undo = NewLink(undo, (long)redo->pos);
    	        }
	        XmTextInsert(w, (XmTextPosition)redo->pos, redo->buffer);
	        undo->length = strlen(redo->buffer);
	    }
	    undo->buffer = XtNewString(redo->buffer);
	    XtFree(redo->buffer);
	    pos = (XmTextPosition)redo->pos;
	    redo = DeleteCurrentLink(redo);
	    if (!redo) {
	        XtSetSensitive(redo_button, False);
	        XtSetSensitive(popup_redo_button, False);
	    }
            XmTextEnableRedisplay(w);
	    XmTextSetInsertionPosition(w, pos);

	    ignoreModify = False;
            if (undo_after_save && file_saved) {
                file_saved = False;
                update_EnvironmentLine();
            }
      	}

	pos = XmTextGetInsertionPosition(w);
	update_StatusLine(w, pos);
}

#ifdef _NO_PROTO
void process_Ctrl_K(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void process_Ctrl_K(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XEvent	theEvent;
	char	kbuf[2];
	static int     length;
	KeySym	key;
	XComposeStatus	cs;
	static XmTextPosition begin, end;

	do {
	    XNextEvent(event->xkey.display, &theEvent);
	    if (theEvent.type == KeyPress) {
	        length = XLookupString(&(theEvent.xkey), kbuf, 2, &key, &cs);
	        kbuf[length] = '\0';
	    }
	} while (theEvent.type != KeyPress && theEvent.type != ButtonPress &&
		theEvent.type != FocusOut);

	switch (key) {
	    case XK_B : case XK_b :	/* <Ctrl><K><B> Mark block-begin */
		begin = XmTextGetInsertionPosition(w);
		break;
	    case XK_K : case XK_k :	/* <Ctrl><K><K> Mark block-end*/
		end = XmTextGetInsertionPosition(w);
		XmTextSetSelection(w, begin, end, theEvent.xkey.time);
		break;
	    case XK_H : case XK_h :	/* <Ctrl><K><H> Hide block */
		XmTextClearSelection(w, theEvent.xkey.time);
		if (found)
		  found = False;
		break;
	    case XK_C : case XK_c :	/* <Ctrl><K><C> Copy Block */
		if (XmTextCopy(w, theEvent.xkey.time) == True)
		  XmTextClearSelection(w, theEvent.xkey.time);
		break;
	    case XK_Y : case XK_y :	/* <Ctrl><K><Y> Delete Block */
		if (XmTextRemove(w) == True)
		  XFlush(event->xkey.display);
		break;
	    case XK_V : case XK_v :	/* <Ctrl><K><V> Paste Block */
		XmTextClearSelection(w, theEvent.xkey.time);
		if (XmTextPaste(w) == True)
		  XFlush(event->xkey.display);
		break;
	    case XK_X : case XK_x :	/* <Ctrl><K><X> Cut Block */
		if (XmTextCut(w, theEvent.xkey.time) == True)
		  XFlush(event->xkey.display);
		break;
	    case XK_I : case XK_i :       /* <Ctrl><K><I> Block Indent */
	        show_cursor_position = False;
                shiftBlock(w, SHIFT_RIGHT, BlockIndentSize, TabSize);
                show_cursor_position = True;
                update_StatusLine(w, XmTextGetInsertionPosition(w));
	        break;
	    case XK_U : case XK_u :       /* <Ctrl><K><U> Block Unindent */
	        show_cursor_position = False;
                shiftBlock(w, SHIFT_LEFT, BlockIndentSize, TabSize);
                show_cursor_position = True;
                update_StatusLine(w, XmTextGetInsertionPosition(w));
	        break;
	    case XK_S : case XK_s :	/* <Ctrl><K><S> Save FIle */
		if (strcmp(FileName, "(NONE)")) {
	            if (!save_file(FileName)) {
		        set_ErrorDialog_message(ErrorDialog, FileName, FILE_SAVE);
		        XtManageChild(ErrorDialog);
	            }
		}
		else {
		    SaveNum = SAVE_DEFAULT;
		    update_SaveAsDialog_data(SaveAsDialog, FileName);
		    XtManageChild(SaveAsDialog);
		}
		break;
	    case XK_P : case XK_p :	/* Upper Case */
	    	UpperCase(text_w);
	    	break;
	    case XK_L : case XK_l :	/* Lower Case */
	    	LowerCase(text_w);
	    	break;
	    case XK_0 : case XK_1 : case XK_2 : case XK_3 : case XK_4 :
	    case XK_5 : case XK_6 : case XK_7 : case XK_8 : case XK_9 :
		place_marker[atoi(kbuf)] = XmTextGetInsertionPosition(w);
		break;		/* Set Place Marker */
	}
}

#ifdef _NO_PROTO
void process_Ctrl_Q(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void process_Ctrl_Q(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XEvent	theEvent;
	char    kbuf[2];
	int     length;
	KeySym	key;
	XComposeStatus	cs;
        XmTextPosition pos, i;
        int page;
	int status;
	Boolean found_str;

	do {
	    XNextEvent(event->xkey.display, &theEvent);
	    if (theEvent.type == KeyPress) {
	        length = XLookupString(&(theEvent.xkey), kbuf, 2, &key, &cs);
	        kbuf[length] = '\0';
	    }
	} while (theEvent.type != KeyPress && theEvent.type != ButtonPress &&
		theEvent.type != FocusOut);

	switch (key) {
	    case XK_Y : case XK_y :	/* <Ctrl><Q><Y> Delete end of line */
		XtCallActionProc(w, "delete-to-end-of-line", event, (String *)NULL, 0);
		break;
	    case XK_P : case XK_p :	/* goto the last cursor Position */
		smTextSetInsertionPosition(w, last_position);
		break;
	    case XK_B : case XK_b :	/* Toggle Create Backup File */
		create_backup = !create_backup;
		break;
	    case XK_S : case XK_s :	/* Beginning of line */
		XtCallActionProc(w, "beginning-of-line", event, (String *)NULL, 0);
		break;
	    case XK_D : case XK_d :	/* End of line */
		XtCallActionProc(w, "end-of-line", event, (String *)NULL, 0);
		break;
	    case XK_E : case XK_e :	/* Top of window */
	  	pos = XmTextGetTopCharacter(w);
		XmTextSetInsertionPosition(w, pos);
		break;
            case XK_X : case XK_x :	/* Bottom of window */
          	pos = smTextGetBottomCharacter(w);
                XmTextSetInsertionPosition(w, pos);
                break;
            case XK_M : case XK_m :     /* Mid of window */
		i = 0;
          	pos = XmTextGetTopCharacter(w);
                page = smTextGetRows(w) >> 1;
        	do {
          	    found_str = XmTextFindString(w, pos, "\n", XmTEXT_FORWARD, &pos);
          	    if (found_str) {
            	        i++;
            	        pos++;
          	    } /* if */
        	} while (found_str && i < page);
        	XmTextSetInsertionPosition(w, pos);
                break;
            case XK_R : case XK_r :       /* Beginning of file */
		XtCallActionProc(w, "beginning-of-file", event, (String *)NULL, 0);
                break;
            case XK_C : case XK_c :       /* End of file */
                XtCallActionProc(w, "end-of-file", event, (String *)NULL, 0);
                break;
	     case XK_bracketleft  : case XK_braceleft  :
	     case XK_bracketright : case XK_braceright :
		i = XmTextGetInsertionPosition(w);
		status = search_pair(w, i, &pos);
		if (status == 0)
		    XmTextSetInsertionPosition(w, pos);
		break;		/*	pair matching  {}()[]	*/
	    case XK_F : case XK_f :
                smManageReplaceText(False);
		if (!XtIsManaged(SearchDialog)) {
          	    SelectionToFind(text_w);
          	    XtManageChild(SearchDialog);
          	}
          	else SelectionToFind(text_w);
	        break;		/* Manage Find Dialog */
	    case XK_A : case XK_a :
                smManageReplaceText(True);
		if (!XtIsManaged(SearchDialog)) {
          	    SelectionToFind(text_w);
          	    XtManageChild(SearchDialog);
          	}
          	else SelectionToFind(text_w);
	        break;		/* Manage Find & Replace Dialog */
	    case XK_0 : case XK_1 : case XK_2 : case XK_3 : case XK_4 :
	    case XK_5 : case XK_6 : case XK_7 : case XK_8 : case XK_9 :
                smTextSetInsertionPosition(w, place_marker[atoi(kbuf)]);
		break;		/* Find Place Marker */
	} /* switch */
	XFlush(event->xkey.display);
}

#ifdef _NO_PROTO
void process_Ctrl_O(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void process_Ctrl_O(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XEvent	theEvent;
	char    kbuf[2];
	static int     length;
	KeySym	key;
	XComposeStatus	cs;

	do {
	    XNextEvent(event->xkey.display, &theEvent);
	    if (theEvent.type == KeyPress) {
	        length = XLookupString(&(theEvent.xkey), kbuf, 2, &key, &cs);
	        kbuf[length] = '\0';
	    }
	} while (theEvent.type != KeyPress && theEvent.type != ButtonPress &&
		theEvent.type != FocusOut);

	switch (key) {
	    case XK_O : case XK_o :	/* Toggle Insert Mode */
	        insert_mode = !insert_mode;
	        XmToggleButtonGadgetSetState(XtNameToWidget(option_env,
	                "button_0"), insert_mode, True);
	        break;
	    case XK_I : case XK_i :	/* Toggle Indent Mode */
	        indent_mode = !indent_mode;
	        XmToggleButtonGadgetSetState(XtNameToWidget(option_env,
                        "button_1"), indent_mode, True);
	        break;
	    case XK_T : case XK_t :	/* Toggle Tab Mode */
	        fill_mode = !fill_mode;
	        XmToggleButtonGadgetSetState(XtNameToWidget(option_env,
	                "button_2"), fill_mode, True);
	        break;
	    case XK_U : case XK_u :	/* Toggle undo after save */
	        undo_after_save = !undo_after_save;
	        XmToggleButtonGadgetSetState(XtNameToWidget(option_env,
	                "button_3"), undo_after_save, True);
	        break;
	    case XK_B : case XK_b :	/* Toggle Create Backup File */
	        create_backup = !create_backup;
	        XmToggleButtonGadgetSetState(XtNameToWidget(option_pref,
	                "button_0"), create_backup, True);
	        break;
	}
	XFlush(event->xkey.display);
}

#ifdef _NO_PROTO
void process_tab(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void process_tab(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	char tabstr[MAXTAB];
	XmTextPosition pos, current;
	int tab, i;
	XmTextPosition index;
	int status;
	char buf[3];

	strcpy(tabstr, "");
	current = pos = XmTextGetInsertionPosition(w);

	index = pos;
	do {
	    index--;
	    status = XmTextGetSubstring(w, index, 2, 3, buf);
	    if (status != XmCOPY_FAILED) {
	        if (buf[0] == '\t')
	            pos -= 8L;
	        else if (buf[0] != '\n')
	            pos--;
	    }
	} while (index >= 0 && buf[0] != '\n' && buf[0] != '\t');
	tab = TabSize - (current - pos) % TabSize;
	for (i = 0; i < tab; i++)
	    strcat(tabstr, " ");
	XmTextInsert(w, current, tabstr);
	XmTextSetInsertionPosition(w, current + (XmTextPosition)tab);
	XFlush(event->xkey.display);
}

#ifdef _NO_PROTO
void BlockIndent(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void BlockIndent(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
        show_cursor_position = False;
        shiftBlock(w, SHIFT_RIGHT, BlockIndentSize, TabSize);
        show_cursor_position = True;
        update_StatusLine(w, XmTextGetInsertionPosition(w));
}

#ifdef _NO_PROTO
void BlockUnindent(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void BlockUnindent(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
        show_cursor_position = False;
        shiftBlock(w, SHIFT_LEFT, BlockIndentSize, TabSize);
        show_cursor_position = True;
        update_StatusLine(w, XmTextGetInsertionPosition(w));
}

#ifdef _NO_PROTO
void move_top(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void move_top(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition pos;

	pos = XmTextGetTopCharacter(w);
	XmTextSetInsertionPosition(w, pos);
}

#ifdef _NO_PROTO
void move_bottom(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void move_bottom(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition pos;

	pos = smTextGetBottomCharacter(w);
	XmTextSetInsertionPosition(w, pos);
}

#ifdef _NO_PROTO
void delete_line(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void delete_line(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition	pos, s_pos, e_pos;
	Boolean found_CR;

	pos = XmTextGetInsertionPosition(w);
	found_CR = XmTextFindString(w, pos - 1, "\n", XmTEXT_BACKWARD, &s_pos);
	if (found_CR) s_pos++;
	else s_pos = 0;
	found_CR = XmTextFindString(w, pos, "\n", XmTEXT_FORWARD, &e_pos);
	if (found_CR) e_pos++;
	else e_pos = XmTextGetLastPosition(w);
	XtRemoveCallback(w, XmNmotionVerifyCallback, Motion_cb, (XtPointer)NULL);
	XmTextSetSelection(w, s_pos, e_pos, event->xkey.time);
	if (XmTextRemove(w) == True)
	    XFlush(event->xkey.display);
	XtAddCallback(w, XmNmotionVerifyCallback, Motion_cb, (XtPointer)NULL);
}

#if (XmVersion >= 1002)	/* Motif 1.2 */
#ifdef _NO_PROTO
void delete_next_word(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void delete_next_word(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
        XmTextPosition start, pos;
        char buf[3];
        int status;

        start = pos = XmTextGetInsertionPosition(w);
        status = XmTextGetSubstring(w, pos, 2, 3, buf);
	if (status != XmCOPY_FAILED && buf[0] == '\n') pos++;
	else if (status != XmCOPY_FAILED && ispunct(buf[0])) pos++;
        else while (status != XmCOPY_FAILED && !ispunct(buf[0]) && !isspace(buf[0])) {
            status = XmTextGetSubstring(w, pos, 2, 3, buf);
            if (status != XmCOPY_FAILED && !ispunct(buf[0]) && !isspace(buf[0]))
                pos++;
        }
        do {
            status = XmTextGetSubstring(w, pos, 2, 3, buf);
            if (status != XmCOPY_FAILED && is_space(buf[0]))
                pos++;
        } while (status != XmCOPY_FAILED && is_space(buf[0]));
	if (start != pos)
	    XmTextReplace(w, start, pos, "");
	XFlush(event->xkey.display);
}
#else
#ifdef _NO_PROTO
int del_next_word(source)
char *source;
#else         /* _NO_PROTO */
int del_next_word(char *source)
#endif
{
	char *str = source;

	if (*str == '\n') str++;
	else if (ispunct(*str)) str++;
        else while (!(ispunct(*str) || isspace(*str))) str++;
        while (is_space(*str))
	    str++;

	return (str - source);
}

#ifdef _NO_PROTO
void delete_next_word(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void delete_next_word(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition	start, pos;

	if (STR == NULL)
	    STR = XmTextGetString(w);
	start = XmTextGetInsertionPosition(w);

	pos = del_next_word(STR + start);
	XmTextReplace(w, start, start + pos, "");
	XFlush(event->xkey.display);
}
#endif

#if (XmVersion >= 1002)	/* Motif 1.2 */
#ifdef _NO_PROTO
void delete_previous_word(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void delete_previous_word(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
        XmTextPosition start, pos;
        char buf[3];
        int status;

        start = pos = XmTextGetInsertionPosition(w);
        status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
	if (status != XmCOPY_FAILED && buf[0] == '\n') pos--;
	else if (status != XmCOPY_FAILED && ispunct(buf[0])) pos--;
        else while (status != XmCOPY_FAILED && !ispunct(buf[0]) && !isspace(buf[0])) {
            status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
            if (status != XmCOPY_FAILED && !ispunct(buf[0]) && !isspace(buf[0]))
                pos--;
        }
        do {
            status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
            if (status != XmCOPY_FAILED && is_space(buf[0]))
                pos--;
        } while (status != XmCOPY_FAILED && is_space(buf[0]));
	if (start != pos)
	    XmTextReplace(w, pos, start, "");
	XFlush(event->xkey.display);
}
#else
#ifdef _NO_PROTO
int del_prev_word(source)
char *source;
#else         /* _NO_PROTO */
int del_prev_word(char *source)
#endif
{
	char *str = source;

	if (*(str - 1) == '\n') str--;
	else if (ispunct(*(str - 1))) str--;
        else while (!(ispunct(*(str - 1)) || isspace(*(str - 1)))) str--;
        while (is_space(*(str - 1)))
	    str--;
	return (source - str);
}

#ifdef _NO_PROTO
void delete_previous_word(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void delete_previous_word(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition start, pos;

	if (STR == NULL)
	    STR = XmTextGetString(w);
	start = XmTextGetInsertionPosition(w);

	pos = del_prev_word(STR + start);
	XmTextReplace(w, start - pos, start, "");
	XFlush(event->xkey.display);
}
#endif

#if (XmVersion >= 1002)	/* Motif 1.2 */
#ifdef _NO_PROTO
void word_right(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else	/* _NO_PROTO */
void word_right(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
        XmTextPosition pos;
        char buf[3];
        int status;

        pos = XmTextGetInsertionPosition(w);
        status = XmTextGetSubstring(w, pos, 2, 3, buf);
        if (status != XmCOPY_FAILED) {
            if (isspace(buf[0]))
	        while (status != XmCOPY_FAILED && isspace(buf[0])) {
	            status = XmTextGetSubstring(w, pos, 2, 3, buf);
	            if (status != XmCOPY_FAILED && isspace(buf[0]))
	                pos++;
	        }
	    else
                while (!isspace(buf[0]) && !ispunct(buf[0])) {
	            status = XmTextGetSubstring(w, pos, 2, 3, buf);
	            if (status != XmCOPY_FAILED && !isspace(buf[0]) && !ispunct(buf[0]))
	                pos++;
                }
        }
        do {
	    status = XmTextGetSubstring(w, pos, 2, 3, buf);
	    if (status != XmCOPY_FAILED && (is_space(buf[0]) || ispunct(buf[0])))
	        pos++;
        } while (status != XmCOPY_FAILED && (is_space(buf[0]) || ispunct(buf[0])));
	XmTextSetInsertionPosition(w, pos);
	XFlush(event->xkey.display);
}
#else
#ifdef _NO_PROTO
int next_word(source)
char *source;
#else	/* _NO_PROTO */
int next_word(char *source)
#endif
{
	char *str = source;

        if (isspace(*str))
	    while (isspace(*str)) str++;
	else
            while (!(isspace(*str) || ispunct(*str))) str++;
        while ((is_space(*str) || ispunct(*str))) str++;

	return (str - source);
}

#ifdef _NO_PROTO
void word_right(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else	/* _NO_PROTO */
void word_right(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition  pos, i;

	if (STR == NULL)
	    STR = XmTextGetString(w);
	i = XmTextGetInsertionPosition(w);
        pos = next_word(STR + i);
	XmTextSetInsertionPosition(w, i + pos);
	XFlush(event->xkey.display);
}
#endif

#if (XmVersion >= 1002)	/* Motif 1.2 */
#ifdef _NO_PROTO
void word_left(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void word_left(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
        XmTextPosition pos;
        char buf[3];
        int status;

        pos = XmTextGetInsertionPosition(w);
        do {
	    status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
	    if (status != XmCOPY_FAILED && (is_space(buf[0]) || ispunct(buf[0])))
	        pos--;
        } while (status != XmCOPY_FAILED && (is_space(buf[0]) || ispunct(buf[0])));
        status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
        if (status != XmCOPY_FAILED) {
            if (isspace(buf[0]))
	        while (status != XmCOPY_FAILED && isspace(buf[0])) {
	            status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
	            if (status != XmCOPY_FAILED && isspace(buf[0]))
	                pos--;
	        }
	    else
                while (!isspace(buf[0]) && !ispunct(buf[0])) {
	            status = XmTextGetSubstring(w, pos - 1, 2, 3, buf);
	            if (status != XmCOPY_FAILED && !isspace(buf[0]) && !ispunct(buf[0]))
	                pos--;
                }
        }
	XmTextSetInsertionPosition(w, pos);
	XFlush(event->xkey.display);

}
#else
#ifdef _NO_PROTO
int prev_word(source)
char *source;
#else	/* _NO_PROTO */
int prev_word(char *source)
#endif
{
	char *str = source;

        while ((is_space(*(str - 1)) || ispunct(*(str - 1)))) str--;
	if (isspace(*(str - 1)))
	    while (isspace(*(str - 1))) str--;
	else
            while (!(isspace(*(str - 1)) || ispunct(*(str - 1)))) str--;

	return (source - str);
}

#ifdef _NO_PROTO
void word_left(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void word_left(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	XmTextPosition i, pos;

	if (STR == NULL)
	    STR = XmTextGetString(w);
	i = XmTextGetInsertionPosition(w);
	pos = prev_word(STR + i);
	XmTextSetInsertionPosition(w, i - pos);
	XFlush(event->xkey.display);
}
#endif

#ifdef _NO_PROTO
void show_info(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
#else
void show_info(Widget w, XEvent *event, String *params, Cardinal *num)
#endif
{
	char buf[BUFSIZ];

	(void)sprintf(buf, "[%ld bytes]", XmTextGetLastPosition(w));
	if (!access(FileName, F_OK) && access(FileName, W_OK) < 0)
	    strcat(buf, " [Read Only]");

	XmTextSetString(status_line, buf);
	XFlush(event->xkey.display);
}

#ifdef _NO_PROTO
void CloseBox(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void CloseBox(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        if (OpenName != NULL) {
            XtFree(OpenName);
            OpenName = NULL;
        }
#ifdef _DEBUG
        printf("Close Box\n");
#endif
	XtUnmanageChild(w);
}

#ifdef _NO_PROTO
void CloseSelectionBox(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void CloseSelectionBox(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
	XtUnmanageChild(w);
}

#ifdef _NO_PROTO
void go_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void go_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmSelectionBoxCallbackStruct *cbs =
			(XmSelectionBoxCallbackStruct *)call_data;
	char *line_str;
	int line;

	XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &line_str);
	if (is_num(line_str)) {
	    line = atoi(line_str);
	    smTextSetInsertionLine(text_w, line);
	}
	XtFree(line_str);
	XFlush(cbs->event->xany.display);
}

#ifdef _NO_PROTO
void settab_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void settab_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmSelectionBoxCallbackStruct *cbs =
                        (XmSelectionBoxCallbackStruct *)call_data;
	char *Tab_Value;

	XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &Tab_Value);
	TabSize = atoi(Tab_Value);
	XtFree(Tab_Value);
	if (TabSize > MAXTAB || TabSize <= 0)
	    TabSize = 8;

        SetTabSize(text_w, TabSize);

	ignoreModify = True;

        XmTextInsert(text_w, 0, "\n");
        XmTextReplace(text_w, 0, 1, "");

	ignoreModify = False;

	XFlush(cbs->event->xany.display);
}

#ifdef _NO_PROTO
void setblock_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void setblock_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmSelectionBoxCallbackStruct *cbs =
                        (XmSelectionBoxCallbackStruct *)call_data;
	char *Value;

	XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &Value);
	BlockIndentSize = atoi(Value);
	XtFree(Value);
	if (BlockIndentSize > MAXTAB || BlockIndentSize <= 0)
	    BlockIndentSize = 2;

	XFlush(cbs->event->xany.display);
}

#ifdef _NO_PROTO
void clearShowMatch(client_data, id)
XtPointer client_data;
XtIntervalId *id;
#else
void clearShowMatch(XtPointer client_data, XtIntervalId *id)
#endif
{
        Widget w = (Widget)client_data;

        XmTextClearSelection(w, CurrentTime);
}

#ifdef _NO_PROTO
void Modify_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void Modify_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmTextVerifyCallbackStruct *cbs =
                        (XmTextVerifyCallbackStruct *)call_data;
	XmTextPosition start, end;
	Boolean is_selected;
	int buffer_size;
#if (XmVersion >= 1002) /* Motif 1.2 */
	int status;
#endif

	if (ignoreModify == False) {
	    if (found)
	        found = False;

	    if (XmTextGetLastPosition(status_line) > 0)
	        XmTextSetString(status_line, "");

            if (redo) {     /* Delete Redo Buffers */
	        while (redo) {
	            if (redo->buffer) XtFree(redo->buffer);
	            redo = DeleteCurrentLink(redo);
	        }
	        XtSetSensitive(redo_button, False);
	        XtSetSensitive(popup_redo_button, False);
            }
	    if (!undo) {
	        XtSetSensitive(undo_button, True);
	        XtSetSensitive(popup_undo_button, True);
	    }

	    if (file_saved) {
	        file_saved = False;
	        update_EnvironmentLine();
	    }

	    undo = NewLink(undo, (long)cbs->startPos);  /* Add New Undo Buffer */

	    if (cbs->text->length > 0) {
	        is_selected = XmTextGetSelectionPosition(w, &start, &end);
                if ((is_selected && start != end) || ReplaceText || !insert_mode) {
 	            buffer_size = (int)(cbs->endPos - cbs->startPos);
	            undo->buffer = XtMalloc((Cardinal)buffer_size + 1);
                    undo->length = -1;
	            status = XmTextGetSubstring(w, cbs->startPos,
	  		      buffer_size, buffer_size + 1, undo->buffer);
	            if (status == XmCOPY_SUCCEEDED)
	                undo->buffer[buffer_size] = '\0';
	            else XtWarning("Insufficient Undo Buffer");
	            undo = NewLink(undo, (long)cbs->startPos);
                }
                undo->length = (long)cbs->text->length;
                undo->buffer = XtNewString(cbs->text->ptr);
                undo->buffer[cbs->text->length] = '\0';
                if (cbs->text->length == 1 && show_match &&
                        strchr("{}()[]<>", cbs->text->ptr[0])) {
                    status = find_matching(w, cbs->startPos, cbs->text->ptr, &start);
                    if (status == 0) {
                        if (start > -1) {
                            XmTextSetSelection(w, start, start + 1, CurrentTime);
                            (void)XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                    1000,  clearShowMatch, (XtPointer)w);
                            smTextSetInsertionPosition(w, cbs->startPos);
                        }
                    }
                    else if (strchr("})]>", cbs->text->ptr[0])) {
                        XBell(cbs->event->xkey.display, 100);
                    }
                }
	    }
	    else { /* cbs->text->length <= 0 */
 	        buffer_size = cbs->endPos - cbs->startPos;
	        undo->buffer = XtMalloc((Cardinal)buffer_size + 1);
	        status = XmTextGetSubstring(w, cbs->startPos,
	  		    buffer_size, buffer_size + 1, undo->buffer);
	        if (status == XmCOPY_SUCCEEDED)
	            undo->buffer[buffer_size] = '\0';
	        else XtWarning("Insufficient Undo Buffer");

	    }
	    if (show_cursor_position)
	        update_StatusLine(w, cbs->startPos);
	}
}

#ifdef _NO_PROTO
void Motion_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void Motion_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
	XmTextVerifyCallbackStruct *cbs =
                        (XmTextVerifyCallbackStruct *)call_data;

	if (ignoreModify == False) {
	    last_position = XmTextGetInsertionPosition(w);

	    if (XmTextGetLastPosition(status_line) > 0)
	        XmTextSetString(status_line, "");
            if (show_cursor_position)
                update_StatusLine(w, cbs->newInsert);
	}
}

#ifdef _NO_PROTO
char *FileGetText(fp)
FILE *fp;
#else
char *FileGetText(FILE *fp)
#endif
{
        char *str;
        int num, position = 0;

        str = XtMalloc((Cardinal)BUFSIZ);
        rewind(fp);
        while ((num = read(fileno(fp), str + position, BUFSIZ)) > 0) {
            position += num;
            str = XtRealloc(str, position + BUFSIZ);
        }
        *(str + position) = '\0';
        return (str);
}

#ifdef _NO_PROTO
Boolean open_file(name)
char *name;
#else
Boolean open_file(char *name)
#endif
{
	struct stat statbuf;
	char buf[BUFSIZ];
	char *str;
        int length;
	FILE *fp;

        if (stat(name, &statbuf) < 0 || !S_ISREG(statbuf.st_mode) ||
                (fp = fopen(name, "r")) == NULL)
	    return (False);

       /* Delete Undo & Redo Buffers */
	while (undo) {
	    if (undo->buffer) XtFree(undo->buffer);
	    undo = DeleteCurrentLink(undo);
	}
	XtSetSensitive(undo_button, False);
	XtSetSensitive(popup_undo_button, False);
	while (redo) {
	    if (redo->buffer) XtFree(redo->buffer);
	    redo = DeleteCurrentLink(redo);
	}
	XtSetSensitive(redo_button, False);
	XtSetSensitive(popup_redo_button, False);

	if (found) {
	    XmTextClearSelection(text_w, CurrentTime);
	    found = False;
	}

/*      pos = ftell(fp);
        fseek(fp, 0L, SEEK_END);
        len = ftell(fp);
        fseek(fp, pos, SEEK_SET);*/

        XmTextSetString(status_line, "Loading...");
	length = (int)statbuf.st_size;

	str = FileGetText(fp);
	fclose(fp);

	XmTextSetString(text_w, str);

	XtFree(str);

	(void)sprintf(buf, "Load %ld bytes", (long)XmTextGetLastPosition(text_w));
	if (access(name, W_OK) < 0)
	    strcat(buf, " [Read Only]");
	XmTextSetString(status_line, buf);
	file_saved = True;

	update_StatusLine(text_w, 0);
        update_EnvironmentLine();

	if ((XmTextPosition)length != XmTextGetLastPosition(text_w))
	    XmTextReplace(text_w, (XmTextPosition)length,
		XmTextGetLastPosition(text_w), "");
        XmProcessTraversal(text_w, XmTRAVERSE_HOME);

	return (True);
}

#ifdef _NO_PROTO
Boolean save_file(name)
char *name;
#else
Boolean save_file(char *name)
#endif
{
	struct stat statbuf;
	char *buf;
	char *str, *opt_str;
	FILE *fp;
	int length;

        /* Create Backup File */
        if (!access(name, F_OK | W_OK) && !stat(name, &statbuf) &&
            S_ISREG(statbuf.st_mode) &&
            create_backup && *(name + strlen(name) - 1) != '%') {
            buf = XtMalloc((Cardinal)(strlen(name) * 2) +  10);
	    (void)sprintf(buf, "cp %s %s%%", name, name);
	    system(buf);
	    XtFree(buf);
        }

	if ((fp = fopen(name, "w+")) == NULL)
	    return (False);

	if (!undo_after_save) {
            /* Delete Undo & Redo Buffers */
	    while (undo) {
	        if (undo->buffer) XtFree(undo->buffer);
	        undo = DeleteCurrentLink(undo);
	    }
	    XtSetSensitive(undo_button, False);
	    XtSetSensitive(popup_undo_button, False);
	    while (redo) {
	        if (redo->buffer) XtFree(redo->buffer);
	        redo = DeleteCurrentLink(redo);
	    }
	    XtSetSensitive(redo_button, False);
	    XtSetSensitive(popup_redo_button, False);
	}

	XmTextSetString(status_line, "Saving...");

	str = XmTextGetString(text_w);

	buf = XtMalloc((Cardinal)BUFSIZ);

        if (optimize_text) {
            opt_str = optimize_str(str);
	    length = strlen(opt_str);
            if (fprintf(fp, "%s", opt_str) == EOF)
	        strcpy(buf, "Warning: Cannot save entire file!");
	    else {
	        if (opt_str[length - 1] != '\n')
	            fputc('\n', fp);
	        (void)sprintf(buf, "Saved %d bytes", length);
	    }
	    XtFree(opt_str);
        }
        else {
	    length = strlen(str);
            if (fprintf(fp, "%s", str) == EOF)
	        strcpy(buf, "Warning: Cannot save entire file!");
	    else {
	        if (str[length - 1] != '\n')
	            fputc('\n', fp);
	        (void)sprintf(buf, "Saved %d bytes", length);
	    }
	}
        XtFree(str);

	fclose(fp);
	XmTextSetString(status_line, buf);
	XtFree(buf);

	file_saved = True;
        update_EnvironmentLine();

	return (True);
}

#ifdef _NO_PROTO
void close_text(w)
Widget w;
#else
void close_text(Widget w)
#endif
{
	ignoreModify = True;
	/* Delete Undo & Redo Buffers */
	while (undo) {
	    if (undo->buffer) XtFree(undo->buffer);
	    undo = DeleteCurrentLink(undo);
	}
	XtSetSensitive(undo_button, False);
	XtSetSensitive(popup_undo_button, False);
	while (redo) {
	    if (redo->buffer) XtFree(redo->buffer);
	    redo = DeleteCurrentLink(redo);
	}
	XtSetSensitive(redo_button, False);
	XtSetSensitive(popup_redo_button, False);

	if (found) {
	    XmTextClearSelection(text_w, CurrentTime);
	    found = False;
	}
	XmTextSetString(w, "");
	ignoreModify = False;

	XtFree(FileName);
	FileName = XtNewString("(NONE)");
	set_title(FileName);

	file_saved = True;
	update_EnvironmentLine();
	update_StatusLine(w, 0);

	XmProcessTraversal(text_w, XmTRAVERSE_HOME);
}

void OpenFileProc()
{
	ignoreModify = True;
        if (open_file(OpenName)) {
            XtFree(FileName);
            FileName = XtNewString(OpenName);
            set_title(FileName);
            add_name(FileName);
        }
        else {
            set_ErrorDialog_message(ErrorDialog, OpenName, FILE_OPEN);
            XtManageChild(ErrorDialog);
        }
        XtFree(OpenName);
        OpenName = NULL;
	ignoreModify = False;
}

void QuitProc()
{
        while (undo) {
            if (undo->buffer) XtFree(undo->buffer);
                undo = DeleteCurrentLink(undo);
        }
        while (redo) {
            if (redo->buffer) XtFree(redo->buffer);
                redo = DeleteCurrentLink(redo);
        }
        XtDestroyWidget(toplevel);
        exit(0);

}

#ifdef _NO_PROTO
Boolean SaveAsProc(SaveNum, name)
int SaveNum;
char *name;
#else
Boolean SaveAsProc(int SaveNum, char *name)
#endif
{
	if (save_file(name)) {
	    XtFree(FileName);
	    FileName = XtNewString(name);
	    switch (SaveNum) {
	        case SAVE_DEFAULT :	/*	Default	*/
		    set_title(FileName);
	  	    add_name(FileName);
	  	    break;
	        case SAVE_NEW :		/*	New	*/
		    XmTextSetString(status_line, "");
		    close_text(text_w);
	    	    break;
	        case SAVE_OPEN :	/*	Open	*/
		    XtManageChild(OpenDialog);
		    break;
	        case SAVE_LIST :	/*	List	*/
	            OpenFileProc();
	    	    break;
	        case SAVE_QUIT :	/*	Quit	*/
	            QuitProc();
		    break;
	    }
	    return (True);
	}
	else return (False);
}

#ifdef _NO_PROTO
void OverWrite_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void OverWrite_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XtUnmanageChild(w);
        if (SaveAsProc(SaveNum, OpenName) == False) {
	    set_ErrorDialog_message(ErrorDialog, OpenName, FILE_SAVE);
	    XtManageChild(ErrorDialog);
        }
        XtFree(OpenName);
        OpenName = NULL;
}

#ifdef _NO_PROTO
void SaveAs_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void SaveAs_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmFileSelectionBoxCallbackStruct *cbs =
                          (XmFileSelectionBoxCallbackStruct *)call_data;
	char *tempName;

        XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &tempName);
        XtUnmanageChild(w);
        if (access(tempName, F_OK) == 0) {
            if (OpenName != NULL) {
                XtFree(OpenName);
                OpenName = NULL;
            }
            OpenName = XtNewString(tempName);
            set_OverwriteDialog_message(OverwriteDialog, OpenName);
            XtManageChild(OverwriteDialog);
        }
        else if (SaveAsProc(SaveNum, tempName) == False) {
	    set_ErrorDialog_message(ErrorDialog, tempName, FILE_SAVE);
	    XtManageChild(ErrorDialog);
        }
	XtFree(tempName);
}

#ifdef _NO_PROTO
void OpenDir_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void OpenDir_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmFileSelectionBoxCallbackStruct *cbs =
                          (XmFileSelectionBoxCallbackStruct *)call_data;
	XmString newName, str;
        char *f_name, *path_name;

	Boolean listUpdated;

        if (strcmp(FileName, "(NONE)")) {
            parseFilename(FileName, &path_name, &f_name);

	    if (!strcmp(f_name, ""))
	        str = XmStringCreateSimple(FileName);
	    else str = XmStringCreateSimple(f_name);

	    XtFree(f_name);
	    XtFree(path_name);
	}
	else str = XmStringCreateSimple("UNNAMED");

	XtVaGetValues(w,
		XmNlistUpdated,	&listUpdated,
		NULL);

	if (listUpdated == True) {
            newName = XmStringConcat(cbs->dir, str);
	    XtVaSetValues(w,
		XmNdirSpec,	newName,
		NULL);
	    XmStringFree(newName);
	}
	XmStringFree(str);
}


#ifdef _NO_PROTO
void yes_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void yes_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
	XtUnmanageChild(w);
	if (strcmp(FileName, "(NONE)")) {
	    if (save_file(FileName)) {
	        switch (WarningNum) {
	            case WARNING_NEW :	/*	New	*/
		        XmTextSetString(status_line, "");
		        close_text(text_w);
		        break;
	            case WARNING_OPEN :	/*	Open	*/
		        XtManageChild(OpenDialog);
		        break;
	            case WARNING_LIST :	/*	List...	*/
			OpenFileProc();
		        break;
	            case WARNING_QUIT :	/*	Quit	*/
	                QuitProc();
		        break;
	        }
	    }
	    else {
	        set_ErrorDialog_message(ErrorDialog, FileName, FILE_SAVE);
	        XtManageChild(ErrorDialog);
	    }
	}
	else {
	    SaveNum = WarningNum + 1;
	    update_SaveAsDialog_data(SaveAsDialog, FileName);
	    XtManageChild(SaveAsDialog);
	}
}

#ifdef _NO_PROTO
void no_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void no_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
	XtUnmanageChild(w);
	switch (WarningNum) {
	    case WARNING_NEW :	/*	New	*/
		XmTextSetString(status_line, "");
		close_text(text_w);
	    	break;
	    case WARNING_OPEN :	/*	Open	*/
	  	XtManageChild(OpenDialog);
	  	break;
	    case WARNING_LIST :	/*	List...	*/
		OpenFileProc();
	    	break;
	    case WARNING_QUIT :	/*	Quit	*/
	        QuitProc();
		break;
	}
}

#ifdef _NO_PROTO
void open_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void open_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmFileSelectionBoxCallbackStruct *cbs =
                          (XmFileSelectionBoxCallbackStruct *)call_data;

        if (OpenName != NULL) {
            XtFree(OpenName);
            OpenName = NULL;
        }
	XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &OpenName);
	XtUnmanageChild(w);
        OpenFileProc();
}

#ifdef _NO_PROTO
void FileMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void FileMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int item_no = (int)client_data;

        XmTextSetString(status_line, "");
	switch (item_no) {
	    case 0 :	/*	New	*/
		if (file_saved) close_text(text_w);
		else {
		    WarningNum = WARNING_NEW;
		    set_QuestionDialog_message(QuestionDialog, FileName);
		    XtManageChild(QuestionDialog);
		}
		break;
	    case 1 :	/*	Open	*/
		if (file_saved) XtManageChild(OpenDialog);
		else {
		    WarningNum = WARNING_OPEN;
		    set_QuestionDialog_message(QuestionDialog, FileName);
		    XtManageChild(QuestionDialog);
		}
		break;
	    case 3 :	/*	Save	*/
		if (strcmp(FileName, "(NONE)")) {
	            if (!save_file(FileName)) {
		        set_ErrorDialog_message(ErrorDialog, FileName, FILE_SAVE);
		        XtManageChild(ErrorDialog);
	            }
		}
		else {
		    SaveNum = SAVE_DEFAULT;
		    update_SaveAsDialog_data(SaveAsDialog, FileName);
		    XtManageChild(SaveAsDialog);
		}
		break;
	    case 4 :	/*	Save As...	*/
		update_SaveAsDialog_data(SaveAsDialog, FileName);
		SaveNum = SAVE_DEFAULT;
		XtManageChild(SaveAsDialog);
		break;
	    case 5 :	/*	Quit	*/
		if (file_saved) QuitProc();
	        else {
		    WarningNum = WARNING_QUIT;
		    set_QuestionDialog_message(QuestionDialog, FileName);
		    XtManageChild(QuestionDialog);
		}
		break;
	}
}

#ifdef _NO_PROTO
void EditMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void EditMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int num = (int)client_data;
        XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct *)call_data;

	switch(num) {
	    case 0 : 	/*	Undo	*/
		XtCallActionProc(text_w, "UndoLast", cbs->event, (String *)NULL, 0);
		break;
	    case 1 : 	/*	Redo	*/
		XtCallActionProc(text_w, "RedoAction", cbs->event, (String *)NULL, 0);
	        break;
	    case 2 : 	/*	Cut	*/
		if (XmTextCut(text_w, cbs->event->xbutton.time) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	    case 3 : 	/*	Copy	*/
		if (XmTextCopy(text_w, cbs->event->xbutton.time) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	    case 4 : 	/*	Paste	*/
		if (XmTextPaste(text_w) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	    case 5 : 	/*	Clear	*/
		if (XmTextRemove(text_w) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	    case 6 :	/*	Select All...	*/
                XtCallActionProc(text_w, "select-all", cbs->event, (String *)NULL, 0);
	  	break;
	    case 7 :	/*	Unselect...	*/
	  	XmTextClearSelection(text_w, cbs->event->xbutton.time);
		if (found)
		    found = False;
	  	break;
	}
}

#ifdef _NO_PROTO
void SearchMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void SearchMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int item_no = (int)client_data;

	switch (item_no) {
            case 0 :	/* Search string */
                smManageReplaceText(False);
		if (!XtIsManaged(SearchDialog)) {
          	    SelectionToFind(text_w);
          	    XtManageChild(SearchDialog);
          	}
          	else SelectionToFind(text_w);
          	break;
            case 1 :	/* Search & Replace string */
                smManageReplaceText(True);
		if (!XtIsManaged(SearchDialog)) {
          	    SelectionToFind(text_w);
          	    XtManageChild(SearchDialog);
          	}
          	else SelectionToFind(text_w);
          	break;
            case 2 :	/* Find again */
                findAgain(text_w);
          	break;
	    case 3 :	/* Go to Line number */
	    	XmTextSetSelection(XmSelectionBoxGetChild(go_dialog, XmDIALOG_TEXT),
	    		0,
	    		XmTextGetLastPosition(XmSelectionBoxGetChild(go_dialog, XmDIALOG_TEXT)),
	    		CurrentTime);
		XtManageChild(go_dialog);
		break;
        }
}

void load_option()	/* Load Option Function */
{
	char *init_file, buf[BUFSIZ];
	FILE *fp;

        init_file = XtMalloc((Cardinal)(strlen(getenv("HOME")) + strlen("/.smeditrc")) + 1);
  	(void)sprintf(init_file, "%s/.smeditrc", getenv("HOME"));

  	if ((fp = fopen(init_file, "r")) != NULL) {
  	    if (fscanf(fp, "Insert Mode=%s\n", buf))
  	        insert_mode = stob(buf);
  	    if (fscanf(fp, "Auto indent=%s\n", buf))
  	        indent_mode = stob(buf);
  	    if (fscanf(fp, "Use tab character=%s\n", buf))
  	        fill_mode = stob(buf);
  	    if (fscanf(fp, "Undo after save=%s\n", buf))
  	        undo_after_save = stob(buf);
  	    if (fscanf(fp, "Show match=%s\n", buf))
  	        show_match = stob(buf);
  	    if (fscanf(fp, "Pending delete=%s\n", buf))
  	        pending_delete = stob(buf);

  	    if (fscanf(fp, "Tab size=%d\n", &TabSize)) {
  	        if (TabSize > MAXTAB || TabSize <= 0) TabSize = 8;
  	    }
  	    if (fscanf(fp, "Block indent=%d\n", &BlockIndentSize)) {
  	        if (BlockIndentSize > MAXTAB || BlockIndentSize <= 0)
	            BlockIndentSize = 2;
            }
  	    if (fscanf(fp, "Create backup file=%s\n", buf))
  	        create_backup = stob(buf);
  	    if (fscanf(fp, "Remove whitespace=%s\n", buf))
  	        optimize_text = stob(buf);

  	    fclose(fp);
  	}
}

#ifdef _NO_PROTO
void OptionMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void OptionMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct *)call_data;
        int num = (int)client_data;
	char *buf;
	FILE *fp;

	switch (num) {
	    case 2 :	/*      Save settings...     */
	        buf = XtMalloc((Cardinal)(strlen(getenv("HOME")) + strlen("/.smeditrc")) + 1);
                (void)sprintf(buf, "%s/.smeditrc", getenv("HOME"));
                if ((fp = fopen(buf, "w")) != NULL) {
  	            if (insert_mode)
  	                (void)fprintf(fp, "Insert Mode=True\n");
  	            else (void)fprintf(fp, "Insert Mode=False\n");
  	            if (indent_mode)
  	                (void)fprintf(fp, "Auto indent=True\n");
  	            else (void)fprintf(fp, "Auto indent=False\n");
  	            if (fill_mode)
  	                (void)fprintf(fp, "Use tab character=True\n");
  	            else (void)fprintf(fp, "Use tab character=False\n");
  	            if (undo_after_save)
  	                (void)fprintf(fp, "Undo after save=True\n");
  	            else (void)fprintf(fp, "Undo after save=False\n");
  	            if (show_match)
  	                (void)fprintf(fp, "Show match=True\n");
  	            else (void)fprintf(fp, "Show match=False\n");
  	            if (pending_delete)
  	                (void)fprintf(fp, "Pending delete=True\n");
  	            else (void)fprintf(fp, "Pending delete=False\n");

  	            (void)fprintf(fp, "Tab size=%d\n", TabSize);
  	            (void)fprintf(fp, "Block indent=%d\n", BlockIndentSize);

  	            if (create_backup)
  	                (void)fprintf(fp, "Create backup file=True\n");
  	            else (void)fprintf(fp, "Create backup file=False\n");
  	            if (optimize_text)
  	                (void)fprintf(fp, "Remove whitespace=True");
  	            else (void)fprintf(fp, "Remove whitespace=False");

                    fclose(fp);
                }
                XtFree(buf);
                break;
	    case 3 :	/*      Refresh Screen  */
#if (XmVersion >= 1002)	/* Motif 1.2 */
                XmUpdateDisplay(text_w);
#endif
		ignoreModify = True;
                if (found) found = False;
                XmTextClearSelection(text_w, cbs->event->xbutton.time);

                buf = XmTextGetString(text_w);
                XmTextSetString(text_w, buf);
                update_StatusLine(text_w, 0);
                XtFree(buf);

		ignoreModify = False;
                break;
	}
}

#ifdef _NO_PROTO
void pick_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void pick_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
	XmString label_str;
	char *pTempStr;

        if (OpenName != NULL) {
            XtFree(OpenName);
            OpenName = NULL;
        }
        XtVaGetValues(w,
            XmNlabelString,	        &label_str,
            NULL);
        XmStringGetLtoR(label_str, XmSTRING_DEFAULT_CHARSET, &pTempStr);
        XmStringFree(label_str);
	OpenName = XtNewString(pTempStr + 3);

	if (file_saved) OpenFileProc();
	else {
	    WarningNum = WARNING_LIST;
	    set_QuestionDialog_message(QuestionDialog, FileName);
	    XtManageChild(QuestionDialog);
	}
}

#ifdef _NO_PROTO
void EnvMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void EnvMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int num = (int)client_data;
        char buf[BUFSIZ];
        XmString str;
        XmToggleButtonCallbackStruct *cbs =
                                   (XmToggleButtonCallbackStruct *)call_data;

	switch (num) {
	    case 0 :        /*      Insert Mode     */
                insert_mode = cbs->set;
                XtCallActionProc(text_w, "toggle-overstrike", cbs->event,
                        (String *)NULL, 0);
		update_EnvironmentLine();
	        break;
	    case 1 :      /*      Indent Mode     */
		indent_mode = cbs->set;
		if (indent_mode)
		    XtOverrideTranslations(text_w,
		  		XtParseTranslationTable(set_indent));
		else
		    XtOverrideTranslations(text_w,
		  		XtParseTranslationTable(set_unindent));
		update_EnvironmentLine();
	    	break;
	    case 2 :      /*      Fill Mode       */
	    	fill_mode = cbs->set;
	    	if (fill_mode)
		    XtOverrideTranslations(text_w,
				XtParseTranslationTable(optimal_fill));
	    	else
	      	    XtOverrideTranslations(text_w,
				XtParseTranslationTable(normal_fill));
		update_EnvironmentLine();
	    	break;
	    case 3 :      /*      Undo After Save		*/
	    	undo_after_save = cbs->set;
	    	break;
	    case 4 :      /*      Auto Matching		*/
	    	show_match = cbs->set;
	    	break;
	    case 5 :      /*      Pending delete	*/
	    	pending_delete = cbs->set;
	    	XtVaSetValues(text_w,
	    		XmNpendingDelete,	pending_delete,
	    		NULL);
	    	break;
	    case 6 :      /*      Set Tab Size     */
            	(void)sprintf(buf, "%d", TabSize);
	    	str = XmStringCreateSimple(buf);
	    	XtVaSetValues(tab_dialog,
	    		XmNtextString,	str,
	    		NULL);
	    	XmStringFree(str);
	    	XmTextSetSelection(XmSelectionBoxGetChild(tab_dialog, XmDIALOG_TEXT),
	    		0,
	    		XmTextGetLastPosition(XmSelectionBoxGetChild(tab_dialog, XmDIALOG_TEXT)),
	    		CurrentTime);
	    	XtManageChild(tab_dialog);
	    	break;
	    case 7 :      /*      Set Block Indent Size     */
            	(void)sprintf(buf, "%d", BlockIndentSize);
	    	str = XmStringCreateSimple(buf);
	    	XtVaSetValues(block_dialog,
	    		XmNtextString,	str,
	    		NULL);
	    	XmStringFree(str);
	    	XmTextSetSelection(XmSelectionBoxGetChild(block_dialog, XmDIALOG_TEXT),
	    		0,
	    		XmTextGetLastPosition(XmSelectionBoxGetChild(block_dialog, XmDIALOG_TEXT)),
	    		CurrentTime);
	    	XtManageChild(block_dialog);
	    	break;
	}

}

#ifdef _NO_PROTO
void PrefMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void PrefMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int num = (int)client_data;
        XmToggleButtonCallbackStruct *cbs =
                                   (XmToggleButtonCallbackStruct *)call_data;

	switch (num) {
	    case 0 :      /*      Create Backup File      */
	    	create_backup = cbs->set;
	    	break;
            case 1 :      /* Optimize Text */
            	optimize_text = cbs->set;
            	break;
	}
}

#ifdef _NO_PROTO
void PopupMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void PopupMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int num = (int)client_data;
        XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct *)call_data;

	switch(num) {
	    case 1 :	/*	Redo	*/
		XtCallActionProc(text_w, "RedoAction", cbs->event, (String *)NULL, 0);
		break;
	    case 2 : 	/*	Cut	*/
		if (XmTextCut(text_w, cbs->event->xbutton.time) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	    case 3 : 	/*	Copy	*/
		if (XmTextCopy(text_w, cbs->event->xbutton.time) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	    case 4 : 	/*	Paste	*/
		if (XmTextPaste(text_w) == True)
		  XFlush(cbs->event->xbutton.display);
		break;
	    case 5 : 	/*	Clear	*/
		if (XmTextRemove(text_w) == True)
		    XFlush(cbs->event->xbutton.display);
		break;
	}
}

#ifdef _NO_PROTO
void PopupSubMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void PopupSubMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
        int num = (int)client_data;
        XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct *)call_data;

	switch(num) {
	    case 0 :	/*	Undo Last Edit	*/
		XtCallActionProc(text_w, "UndoLast", cbs->event, (String *)NULL, 0);
		break;
	    case 1 :	/*	Undo All Edits	*/
		XtCallActionProc(text_w, "UndoAll", cbs->event, (String *)NULL, 0);
		break;
	}
}

#ifdef _NO_PROTO
void HelpMenu_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
#else
void HelpMenu_cb(Widget w, XtPointer client_data, XtPointer call_data)
#endif
{
	int item_no = (int)client_data;

	switch (item_no) {
	    case 0 :
		if (XtIsManaged(help_dialog) == False)
		    loadHelpText();
		XtManageChild(help_dialog);
	  	break;
	    case 1 :
		XtManageChild(about_dialog);
	  	break;
	}
}
